from typing import Union, List, Tuple
from numpy import array, ndarray, log
from scipy.stats import rankdata, chi2


def global_simes_test(p_vals: Union[List, ndarray]) -> float:
    '''Global Simes test of the intersection null hypothesis.

    Computes the combined p value as min(np(i)/i), where p(1), ..., p(n) are
    the ordered p values [1]_.

    Parameters
    ----------
    p_vals : Union[List, ndarray]
        An array of p values.

    Returns
    -------
    p_value : float
        Global p value.

    References
    ----------
    .. [1] Simes, R. J. (1986). An improved Bonferroni procedure for multiple
        tests of significance. Biometrika, 73(3):751-754.

    Examples
    --------
    >>> arr = [0.04, 0.03, 0.98, 0.01, 0.43, 0.99, 1.0, 0.002]
    >>> sp.global_simes_test(arr)
    '''
    arr = array(p_vals)
    ranks = rankdata(arr)
    p_value = min(arr.size * arr / ranks)
    return p_value


def global_f_test(
        p_vals: Union[List, ndarray],
        stat: bool = False) -> Union[float, Tuple[float, float]]:
    '''Fisher's combination test for global null hypothesis.

    Computes the combined p value using chi-squared distribution and T
    statistic: -2 * sum(log(x)) [1]_.

    Parameters
    ----------
    p_vals : Union[List, ndarray]
        An array or a list of p values.
    stat : bool
        Defines if statistic should be returned.

    Returns
    -------
    p_value : float
        Global p value.
    t_stat : float
        Statistic.

    References
    ----------
    .. [1] Fisher RA. Statistical methods for research workers,
        London: Oliver and Boyd, 1932.

    Examples
    --------
    >>> x = [0.04, 0.03, 0.98, 0.01, 0.43, 0.99, 1.0, 0.002]
    >>> sp.global_f_test(x)
    '''
    arr = array(p_vals)
    t_stat = -2 * sum(log(arr))
    p_value = chi2.sf(t_stat, df=2 * len(arr))
    return p_value, t_stat if stat else p_value




import itertools as it
from typing import Optional, Union, Literal
import numpy as np
from numpy.typing import ArrayLike
import scipy.stats as ss
from statsmodels.sandbox.stats.multicomp import multipletests
from pandas import DataFrame, Series, MultiIndex


def __convert_to_df(
    a: Union[list, np.ndarray, DataFrame],
    val_col: Optional[str] = "vals",
    group_col: Optional[str] = "groups",
    val_id: Optional[int] = None,
    group_id: Optional[int] = None,
) -> tuple[DataFrame, str, str]:
    """Hidden helper method to create a DataFrame with input data for further
    processing.

    Parameters
    ----------
    a : Union[list, np.ndarray, DataFrame]
        An array, any object exposing array interface or a pandas DataFrame.
        Array must be two-dimensional. Second dimension may vary, i.e. groups
        may have different lengths.

    val_col : str, optional
        Name of a DataFrame column that contains dependent variable values
        (test or response variable). Values should have a non-nominal scale.
        Must be specified if `a` is a pandas DataFrame object.

    group_col : str, optional
        Name of a DataFrame column that contains independent variable values
        (grouping or predictor variable). Values should have a nominal scale
        (categorical). Must be specified if `a` is a pandas DataFrame object.

    val_id : int, optional
        Index of a column that contains dependent variable values (test or
        response variable). Should be specified if a NumPy ndarray is used as an
        input. It will be inferred from data, if not specified.

    group_id : int, optional
        Index of a column that contains independent variable values (grouping or
        predictor variable). Should be specified if a NumPy ndarray is used as
        an input. It will be inferred from data, if not specified.

    Returns
    -------
    tuple[DataFrame, str, str]
        Returns a tuple of DataFrame and two strings:
        - DataFrame with input data, `val_col` column contains numerical values
          and `group_col` column contains categorical values.
        - Name of a DataFrame column that contains dependent variable values
          (test or response variable).
        - Name of a DataFrame column that contains independent variable values
          (grouping or predictor variable).

    Notes
    -----
    Inferrence algorithm for determining `val_id` and `group_id` args is rather
    simple, so it is better to specify them explicitly to prevent errors.
    """
    if not group_col:
        group_col = "groups"
    if not val_col:
        val_col = "vals"

    if isinstance(a, DataFrame):
        x = a.copy()
        if not {group_col, val_col}.issubset(a.columns):
            raise ValueError("Specify correct column names using `group_col` and `val_col` args")
        return x, val_col, group_col

    elif isinstance(a, list) or (isinstance(a, np.ndarray) and not a.shape.count(2)):
        grps_len = map(len, a)
        grps = list(it.chain(*[[i + 1] * grp_len for i, grp_len in enumerate(grps_len)]))
        vals = list(it.chain(*a))

        return DataFrame({val_col: vals, group_col: grps}), val_col, group_col

    elif isinstance(a, np.ndarray):
        # cols ids not defined
        # trying to infer
        if not all([val_id, group_id]):
            if np.argmax(a.shape):
                a = a.T

            ax = [np.unique(a[:, 0]).size, np.unique(a[:, 1]).size]

            if np.diff(ax).item():
                __val_col = np.argmax(ax)
                __group_col = np.argmin(ax)
            else:
                raise ValueError(
                    "Cannot infer input format.\nPlease specify `val_id` and `group_id` args"
                )

            cols = {__val_col: val_col, __group_col: group_col}
        else:
            cols = {val_id: val_col, group_id: group_col}

        cols_vals = np.array(dict(sorted(cols.items())).values())
        return DataFrame(a, columns=cols_vals).dropna(), val_col, group_col


def __convert_to_block_df(
    a: Union[DataFrame, ArrayLike],
    y_col: Optional[Union[str, int]] = None,
    group_col: Optional[Union[str, int]] = None,
    block_col: Optional[Union[str, int]] = None,
    block_id_col: Optional[Union[str, int]] = None,
    melted: bool = False,
) -> tuple[DataFrame, str, str, str, str]:
    if melted and np.any(np.array([block_col, group_col, y_col]) == None):
        raise ValueError(
            "`block_col`, `group_col`, `y_col` should be explicitly specified if using melted data"
        )

    new_block_id_col = "block_ids"
    new_group_col = "groups"
    new_block_col = "blocks"
    new_y_col = "y"

    if isinstance(a, DataFrame) and not melted:
        x = a.copy(deep=True)
        x.columns.name = new_group_col
        x.index.name = new_block_col
        x[new_block_id_col] = np.arange(x.shape[0])
        x = x.reset_index().melt(
            id_vars=[new_block_col, new_block_id_col], var_name=new_group_col, value_name=new_y_col
        )

    elif isinstance(a, DataFrame) and melted:
        if a[block_col].duplicated().sum() > 0 and not block_id_col:
            raise ValueError(
                "`block_col` contains duplicated entries, `block_id_col` should be explicitly specified"
            )

        x = DataFrame.from_dict(
            {
                new_group_col: a[group_col],
                new_block_col: a[block_col],
                new_y_col: a[y_col],
                new_block_id_col: a[block_id_col],
            }
        )

    else:
        x = np.array(a)
        x = DataFrame(x, index=np.arange(x.shape[0]), columns=np.arange(x.shape[1]))

        if not melted:
            x.columns.name = new_group_col
            x.index.name = new_block_col
            x[new_block_id_col] = np.arange(x.shape[0])
            x = x.reset_index().melt(
                id_vars=[new_block_col, new_block_id_col],
                var_name=new_group_col,
                value_name=new_y_col,
            )

        else:
            x.rename(
                columns={
                    group_col: new_group_col,
                    block_col: new_block_col,
                    y_col: new_y_col,
                    block_id_col: new_block_id_col,
                },
                inplace=True,
            )

    return x, new_y_col, new_group_col, new_block_col, new_block_id_col


def posthoc_conover(
    a: Union[list, np.ndarray, DataFrame],
    val_col: Optional[str] = None,
    group_col: Optional[str] = None,
    p_adjust: Optional[str] = None,
    sort: bool = True,
) -> DataFrame:
    """Post hoc pairwise test for multiple comparisons of mean rank sums
    (Conover´s test). May be used after Kruskal-Wallis one-way analysis of
    variance by ranks to do pairwise comparisons [1]_.

    Parameters
    ----------
    a : array_like or pandas DataFrame object
        An array, any object exposing the array interface or a pandas DataFrame.
        Array must be two-dimensional. Second dimension may vary,
        i.e. groups may have different lengths.

    val_col : str, optional
        Name of a DataFrame column that contains dependent variable values (test
        or response variable). Values should have a non-nominal scale. Must be
        specified if `a` is a pandas DataFrame object.

    group_col : str, optional
        Name of a DataFrame column that contains independent variable values
        (grouping or predictor variable). Values should have a nominal scale
        (categorical). Must be specified if `a` is a pandas DataFrame object.

    p_adjust : str, optional
        Method for adjusting p values. See `statsmodels.sandbox.stats.multicomp`
        for details. Available methods are:
        'bonferroni' : one-step correction
        'sidak' : one-step correction
        'holm-sidak' : step-down method using Sidak adjustments
        'holm' : step-down method using Bonferroni adjustments
        'simes-hochberg' : step-up method  (independent)
        'hommel' : closed method based on Simes tests (non-negative)
        'fdr_bh' : Benjamini/Hochberg  (non-negative)
        'fdr_by' : Benjamini/Yekutieli (negative)
        'fdr_tsbh' : two stage fdr correction (non-negative)
        'fdr_tsbky' : two stage fdr correction (non-negative)

    sort : bool, optional
        Specifies whether to sort DataFrame by `group_col` or not. Recommended
        unless you sort your data manually.

    Returns
    -------
    result : pandas.DataFrame
        P values.

    Notes
    -----
    A tie correction are employed according to Conover [1]_.

    References
    ----------
    .. [1] W. J. Conover and R. L. Iman (1979), On multiple-comparisons
        procedures, Tech. Rep. LA-7677-MS, Los Alamos Scientific Laboratory.

    Examples
    --------
    >>> x = [[1,2,3,5,1], [12,31,54, np.nan], [10,12,6,74,11]]
    >>> sp.posthoc_conover(x, p_adjust = 'holm')
    """

    def compare_conover(i, j):
        diff = np.abs(x_ranks_avg.loc[i] - x_ranks_avg.loc[j])
        B = 1.0 / x_lens.loc[i] + 1.0 / x_lens.loc[j]
        D = (n - 1.0 - h_cor) / (n - x_len)
        t_value = diff / np.sqrt(S2 * B * D)
        p_value = 2.0 * ss.t.sf(np.abs(t_value), df=n - x_len)
        return p_value

    x, _val_col, _group_col = __convert_to_df(a, val_col, group_col)
    x = x.sort_values(by=[_group_col, _val_col], ascending=True) if sort else x

    n = len(x.index)
    x_groups_unique = x[_group_col].unique()
    x_len = x_groups_unique.size
    x_lens = x.groupby(_group_col, observed=True)[_val_col].count()

    x["ranks"] = x[_val_col].rank()
    x_ranks_avg = x.groupby(_group_col, observed=True)["ranks"].mean()
    x_ranks_sum = x.groupby(_group_col, observed=True)["ranks"].sum().to_numpy()

    # ties
    vals = x.groupby("ranks").count()[_val_col].to_numpy()
    tie_sum = np.sum(vals[vals != 1] ** 3 - vals[vals != 1])
    tie_sum = 0 if not tie_sum else tie_sum
    x_ties = np.min([1.0, 1.0 - tie_sum / (n**3.0 - n)])

    h = (12.0 / (n * (n + 1.0))) * np.sum(x_ranks_sum**2 / x_lens) - 3.0 * (n + 1.0)
    h_cor = h / x_ties

    if x_ties == 1:
        S2 = n * (n + 1.0) / 12.0
    else:
        S2 = (1.0 / (n - 1.0)) * (np.sum(x["ranks"] ** 2.0) - (n * (((n + 1.0) ** 2.0) / 4.0)))

    vs = np.zeros((x_len, x_len))
    tri_upper = np.triu_indices(vs.shape[0], 1)
    tri_lower = np.tril_indices(vs.shape[0], -1)
    vs[:, :] = 0

    combs = it.combinations(range(x_len), 2)

    for i, j in combs:
        vs[i, j] = compare_conover(x_groups_unique[i], x_groups_unique[j])

    if p_adjust:
        vs[tri_upper] = multipletests(vs[tri_upper], method=p_adjust)[1]
    vs[tri_lower] = np.transpose(vs)[tri_lower]
    np.fill_diagonal(vs, 1)

    return DataFrame(vs, index=x_groups_unique, columns=x_groups_unique)


def posthoc_dunn(
    a: Union[list, np.ndarray, DataFrame],
    val_col: Optional[str] = None,
    group_col: Optional[str] = None,
    p_adjust: Optional[str] = None,
    sort: bool = True,
) -> DataFrame:
    """Post hoc pairwise test for multiple comparisons of mean rank sums
    (Dunn's test). May be used after Kruskal-Wallis one-way analysis of
    variance by ranks to do pairwise comparisons [1]_, [2]_.

    Parameters
    ----------
    a : array_like or pandas DataFrame object
        An array, any object exposing the array interface or a pandas DataFrame.
        Array must be two-dimensional. Second dimension may vary,
        i.e. groups may have different lengths.

    val_col : str, optional
        Name of a DataFrame column that contains dependent variable values (test
        or response variable). Values should have a non-nominal scale. Must be
        specified if `a` is a pandas DataFrame object.

    group_col : str, optional
        Name of a DataFrame column that contains independent variable values
        (grouping or predictor variable). Values should have a nominal scale
        (categorical). Must be specified if `a` is a pandas DataFrame object.

    p_adjust : str, optional
        Method for adjusting p values. See `statsmodels.sandbox.stats.multicomp`
        for details. Available methods are:
        'bonferroni' : one-step correction
        'sidak' : one-step correction
        'holm-sidak' : step-down method using Sidak adjustments
        'holm' : step-down method using Bonferroni adjustments
        'simes-hochberg' : step-up method  (independent)
        'hommel' : closed method based on Simes tests (non-negative)
        'fdr_bh' : Benjamini/Hochberg  (non-negative)
        'fdr_by' : Benjamini/Yekutieli (negative)
        'fdr_tsbh' : two stage fdr correction (non-negative)
        'fdr_tsbky' : two stage fdr correction (non-negative)

    sort : bool, optional
        Specifies whether to sort DataFrame by group_col or not. Recommended
        unless you sort your data manually.

    Returns
    -------
    result : pandas.DataFrame
        P values.

    Notes
    -----
    A tie correction will be employed according to Glantz (2012).

    References
    ----------
    .. [1] O.J. Dunn (1964). Multiple comparisons using rank sums.
        Technometrics, 6, 241-252.
    .. [2] S.A. Glantz (2012), Primer of Biostatistics. New York: McGraw Hill.

    Examples
    --------

    >>> x = [[1,2,3,5,1], [12,31,54, np.nan], [10,12,6,74,11]]
    >>> sp.posthoc_dunn(x, p_adjust = 'holm')
    """

    def compare_dunn(i, j):
        diff = np.abs(x_ranks_avg.loc[i] - x_ranks_avg.loc[j])
        A = n * (n + 1.0) / 12.0
        B = 1.0 / x_lens.loc[i] + 1.0 / x_lens.loc[j]
        z_value = diff / np.sqrt((A - x_ties) * B)
        p_value = 2.0 * ss.norm.sf(np.abs(z_value))
        return p_value

    x, _val_col, _group_col = __convert_to_df(a, val_col, group_col)
    x = x.sort_values(by=[_group_col, _val_col], ascending=True) if sort else x

    n = len(x.index)
    x_groups_unique = x[_group_col].unique()
    x_len = x_groups_unique.size
    x_lens = x.groupby(_group_col, observed=True)[_val_col].count()

    x["ranks"] = x[_val_col].rank()
    x_ranks_avg = x.groupby(_group_col, observed=True)["ranks"].mean()

    # ties
    vals = x.groupby("ranks").count()[_val_col].to_numpy()
    tie_sum = np.sum(vals[vals != 1] ** 3 - vals[vals != 1])
    tie_sum = 0 if not tie_sum else tie_sum
    x_ties = tie_sum / (12.0 * (n - 1))

    vs = np.zeros((x_len, x_len))
    combs = it.combinations(range(x_len), 2)

    tri_upper = np.triu_indices(vs.shape[0], 1)
    tri_lower = np.tril_indices(vs.shape[0], -1)
    vs[:, :] = 0

    for i, j in combs:
        vs[i, j] = compare_dunn(x_groups_unique[i], x_groups_unique[j])

    if p_adjust:
        vs[tri_upper] = multipletests(vs[tri_upper], method=p_adjust)[1]

    vs[tri_lower] = np.transpose(vs)[tri_lower]
    np.fill_diagonal(vs, 1)
    return DataFrame(vs, index=x_groups_unique, columns=x_groups_unique)


def posthoc_nemenyi(
    a: Union[list, np.ndarray, DataFrame],
    val_col: Optional[str] = None,
    group_col: Optional[str] = None,
    dist: str = "chi",
    sort: bool = True,
) -> DataFrame:
    """Post hoc pairwise test for multiple comparisons of mean rank sums
    (Nemenyi's test). May be used after Kruskal-Wallis one-way analysis of
    variance by ranks to do pairwise comparisons [1]_.

    Parameters
    ----------
    a : array_like or pandas DataFrame object
        An array, any object exposing the array interface or a pandas
        DataFrame. Array must be two-dimensional. Second dimension may vary,
        i.e. groups may have different lengths.

    val_col : str, optional
        Name of a DataFrame column that contains dependent variable values (test
        or response variable). Values should have a non-nominal scale. Must be
        specified if `a` is a pandas DataFrame object.

    group_col : str, optional
        Name of a DataFrame column that contains independent variable values
        (grouping or predictor variable). Values should have a nominal scale
        (categorical). Must be specified if `a` is a pandas DataFrame object.

    dist : str, optional
        Method for determining the p value. The default distribution is "chi"
        (chi-squared), else "tukey" (studentized range).

    sort : bool, optional
        Specifies whether to sort DataFrame by group_col or not. Recommended
        unless you sort your data manually.

    Returns
    -------
    result : pandas.DataFrame
        P values.

    Notes
    -----
    A tie correction will be employed according to Glantz (2012).

    References
    ----------
    .. [1] Lothar Sachs (1997), Angewandte Statistik. Berlin: Springer.
        Pages: 395-397, 662-664.

    Examples
    --------
    >>> x = [[1,2,3,5,1], [12,31,54, np.nan], [10,12,6,74,11]]
    >>> sp.posthoc_nemenyi(x)
    """

    def compare_stats_chi(i, j):
        diff = np.abs(x_ranks_avg.loc[i] - x_ranks_avg.loc[j])
        A = n * (n + 1.0) / 12.0
        B = 1.0 / x_lens.loc[i] + 1.0 / x_lens.loc[j]
        chi = diff**2.0 / (A * B)
        return chi

    def compare_stats_tukey(i, j):
        diff = np.abs(x_ranks_avg.loc[i] - x_ranks_avg.loc[j])
        B = 1.0 / x_lens.loc[i] + 1.0 / x_lens.loc[j]
        q = diff / np.sqrt((n * (n + 1.0) / 12.0) * B)
        return q

    x, _val_col, _group_col = __convert_to_df(a, val_col, group_col)
    x = x.sort_values(by=[_group_col, _val_col], ascending=True) if sort else x

    n = len(x.index)
    x_groups_unique = x[_group_col].unique()
    x_len = x_groups_unique.size
    x_lens = x.groupby(_group_col, observed=True)[_val_col].count()

    x["ranks"] = x[_val_col].rank()
    x_ranks_avg = x.groupby(_group_col, observed=True)["ranks"].mean()

    # ties
    vals = x.groupby("ranks").count()[_val_col].to_numpy()
    tie_sum = np.sum(vals[vals != 1] ** 3 - vals[vals != 1])
    tie_sum = 0 if not tie_sum else tie_sum
    x_ties = np.min([1.0, 1.0 - tie_sum / (n**3.0 - n)])

    vs = np.zeros((x_len, x_len))
    combs = it.combinations(range(x_len), 2)

    tri_upper = np.triu_indices(vs.shape[0], 1)
    tri_lower = np.tril_indices(vs.shape[0], -1)
    vs[:, :] = 0

    if dist == "chi":
        for i, j in combs:
            vs[i, j] = compare_stats_chi(x_groups_unique[i], x_groups_unique[j]) / x_ties

        vs[tri_upper] = ss.chi2.sf(vs[tri_upper], x_len - 1)

    elif dist == "tukey":
        for i, j in combs:
            vs[i, j] = compare_stats_tukey(x_groups_unique[i], x_groups_unique[j]) * np.sqrt(2.0)

        vs[tri_upper] = ss.studentized_range.sf(vs[tri_upper], x_len, np.inf)

    vs[tri_lower] = np.transpose(vs)[tri_lower]
    np.fill_diagonal(vs, 1)

    return DataFrame(vs, index=x_groups_unique, columns=x_groups_unique)


def posthoc_nemenyi_friedman(
    a: Union[list, np.ndarray, DataFrame],
    y_col: Optional[str] = None,
    group_col: Optional[str] = None,
    block_col: Optional[str] = None,
    block_id_col: Optional[str] = None,
    melted: bool = False,
    sort: bool = False,
) -> DataFrame:
    """Calculate pairwise comparisons using Nemenyi post hoc test for
    unreplicated blocked data. This test is usually conducted post hoc if
    significant results of the Friedman's test are obtained. The statistics
    refer to upper quantiles of the studentized range distribution (Tukey) [1]_,
    [2]_, [3]_.

    Parameters
    ----------
    a : array_like or pandas DataFrame object
        An array, any object exposing the array interface or a pandas
        DataFrame.

        If `melted` is set to False (default), `a` is a typical matrix of
        block design, i.e. rows are blocks, and columns are groups. In this
        case you do not need to specify col arguments.

        If `a` is an array and `melted` is set to True,
        y_col, block_col and group_col must specify the indices of columns
        containing elements of correspondary type.

        If `a` is a Pandas DataFrame and `melted` is set to True,
        y_col, block_col and group_col must specify columns names (strings).

    y_col : str or int
        Must be specified if `a` is a pandas DataFrame object.
        Name of the column that contains y data.

    block_col : str or int
        Must be specified if `a` is a pandas DataFrame object.
        Name of the column that contains blocking factor values.

    group_col : str or int
        Must be specified if `a` is a pandas DataFrame object.
        Name of the column that contains treatment (group) factor values.

    melted : bool, optional
        Specifies if data are given as melted columns "y", "blocks", and
        "groups".

    sort : bool, optional
        If True, sort data by block and group columns.

    Returns
    -------
    result : pandas.DataFrame
        P values.

    Notes
    -----
    A one-way ANOVA with repeated measures that is also referred to as ANOVA
    with unreplicated block design can also be conducted via Friedman's
    test. The consequent post hoc pairwise multiple comparison test
    according to Nemenyi is conducted with this function.

    This function does not test for ties.

    References
    ----------
    .. [1] J. Demsar (2006), Statistical comparisons of classifiers over
        multiple data sets, Journal of Machine Learning Research, 7, 1-30.

    .. [2] P. Nemenyi (1963) Distribution-free Multiple Comparisons. Ph.D.
        thesis, Princeton University.

    .. [3] L. Sachs (1997), Angewandte Statistik. Berlin: Springer.
        Pages: 668-675.

    Examples
    --------
    >>> # Non-melted case, x is a block design matrix, i.e. rows are blocks
    >>> # and columns are groups.
    >>> x = np.array([[31,27,24],[31,28,31],[45,29,46],[21,18,48],[42,36,46],[32,17,40]])
    >>> sp.posthoc_nemenyi_friedman(x)
    """

    def compare_stats(i, j):
        dif = np.abs(R[groups[i]] - R[groups[j]])
        qval = dif / np.sqrt(k * (k + 1.0) / (6.0 * n))
        return qval

    x, _y_col, _group_col, _block_col, _block_id_col = __convert_to_block_df(
        a, y_col, group_col, block_col, block_id_col, melted
    )
    x = x.sort_values(by=[_group_col, _block_col], ascending=True) if sort else x
    x.dropna(inplace=True)

    groups = x[_group_col].unique()
    k = groups.size
    n = x[_block_id_col].unique().size

    x["mat"] = x.groupby(_block_id_col, observed=True)[_y_col].rank()
    R = x.groupby(_group_col, observed=True)["mat"].mean()
    vs = np.zeros((k, k))
    combs = it.combinations(range(k), 2)

    tri_upper = np.triu_indices(vs.shape[0], 1)
    tri_lower = np.tril_indices(vs.shape[0], -1)
    vs[:, :] = 0

    for i, j in combs:
        vs[i, j] = compare_stats(i, j)

    vs *= np.sqrt(2.0)
    vs[tri_upper] = ss.studentized_range.sf(vs[tri_upper], k, np.inf)
    vs[tri_lower] = np.transpose(vs)[tri_lower]
    np.fill_diagonal(vs, 1)
    return DataFrame(vs, index=groups, columns=groups)


def posthoc_conover_friedman(
    a: Union[list, np.ndarray, DataFrame],
    y_col: Optional[str] = None,
    block_col: Optional[str] = None,
    block_id_col: Optional[str] = None,
    group_col: Optional[str] = None,
    melted: bool = False,
    sort: bool = False,
    p_adjust: Optional[str] = None,
) -> DataFrame:
    """Calculate pairwise comparisons using Conover post hoc test for unreplicated
    blocked data. This test is usually conducted post hoc after
    significant results of the Friedman test. The statistics refer to
    the Student t distribution [1]_, [2]_.

    Parameters
    ----------
    a : array_like or pandas DataFrame object
        An array, any object exposing the array interface or a pandas
        DataFrame.

        If `melted` is set to False (default), `a` is a typical matrix of
        block design, i.e. rows are blocks, and columns are groups. In this
        case you do not need to specify col arguments.

        If `a` is an array and `melted` is set to True,
        y_col, block_col and group_col must specify the indices of columns
        containing elements of correspondary type.

        If `a` is a Pandas DataFrame and `melted` is set to True,
        y_col, block_col and group_col must specify columns names (strings).

    y_col : str or int
        Must be specified if `a` is a pandas DataFrame object.
        Name of the column that contains y data.

    block_col : str or int
        Must be specified if `a` is a melted pandas DataFrame object.
        Name of the column that contains blocking factor values.

    block_id_col : str or int
        Must be specified if `a` is a melted pandas DataFrame object.
        Name of the column that contains identifiers of blocking factor values.

    group_col : str or int
        Must be specified if `a` is a melted pandas DataFrame object.
        Name of the column that contains treatment (group) factor values.

    melted : bool, optional
        Specifies if data are given as melted columns "y", "blocks", and
        "groups".

    sort : bool, optional
        If True, sort data by block and group columns.

    p_adjust : str, optional
        Method for adjusting p values. See statsmodels.sandbox.stats.multicomp
        for details. Available methods are:
        'bonferroni' : one-step correction
        'sidak' : one-step correction
        'holm-sidak' : step-down method using Sidak adjustments
        'holm' : step-down method using Bonferroni adjustments
        'simes-hochberg' : step-up method  (independent)
        'hommel' : closed method based on Simes tests (non-negative)
        'fdr_bh' : Benjamini/Hochberg  (non-negative)
        'fdr_by' : Benjamini/Yekutieli (negative)
        'fdr_tsbh' : two stage fdr correction (non-negative)
        'fdr_tsbky' : two stage fdr correction (non-negative)
        'single-step' : uses Tukey distribution for multiple comparisons

    Returns
    -------
    result : pandas.DataFrame
        P values.

    Notes
    -----
    A one-way ANOVA with repeated measures that is also referred to as ANOVA
    with unreplicated block design can also be conducted via the
    friedman.test. The consequent post hoc pairwise multiple comparison test
    according to Conover is conducted with this function.

    If y is a matrix, than the columns refer to the treatment and the rows
    indicate the block.

    References
    ----------
    .. [1] W. J. Conover and R. L. Iman (1979), On multiple-comparisons procedures,
        Tech. Rep. LA-7677-MS, Los Alamos Scientific Laboratory.

    .. [2] W. J. Conover (1999), Practical nonparametric Statistics, 3rd. Edition,
        Wiley.

    Examples
    --------
    >>> x = np.array([[31,27,24],[31,28,31],[45,29,46],[21,18,48],[42,36,46],[32,17,40]])
    >>> sp.posthoc_conover_friedman(x)
    """

    def compare_stats(i, j):
        dif = np.abs(R.loc[groups[i]] - R.loc[groups[j]])
        tval = dif / np.sqrt(A) / np.sqrt(B)
        pval = 2.0 * ss.t.sf(np.abs(tval), df=(m * n * k - k - n + 1)).item()
        return pval

    def compare_tukey(i, j):
        dif = np.abs(R.loc[groups[i]] - R.loc[groups[j]])
        qval = np.sqrt(2.0) * dif / np.sqrt(A) / np.sqrt(B)
        pval = ss.studentized_range.sf(qval, k, np.inf).item()
        return pval

    x, _y_col, _group_col, _block_col, _block_id_col = __convert_to_block_df(
        a, y_col, group_col, block_col, block_id_col, melted
    )
    x = x.sort_values(by=[_group_col, _block_col], ascending=True) if sort else x
    x.dropna(inplace=True)

    groups = x[_group_col].unique()
    k = groups.size
    n = x[_block_id_col].unique().size

    x["mat"] = x.groupby(_block_id_col, observed=True)[_y_col].rank()
    R = x.groupby(_group_col, observed=True)["mat"].sum()
    A1 = (x["mat"] ** 2).sum()
    m = 1
    S2 = m / (m * k - 1.0) * (A1 - m * k * n * ((m * k + 1.0) ** 2.0) / 4.0)
    T2 = 1.0 / S2 * (np.sum((R.to_numpy() - n * m * (m * k + 1.0) / 2.0) ** 2.0))
    A = S2 * (2.0 * n * (m * k - 1.0)) / (m * n * k - k - n + 1.0)
    B = 1.0 - T2 / (n * (m * k - 1.0))

    vs = np.zeros((k, k))
    combs = it.combinations(range(k), 2)

    tri_upper = np.triu_indices(vs.shape[0], 1)
    tri_lower = np.tril_indices(vs.shape[0], -1)
    vs[:, :] = 0

    if p_adjust == "single-step":
        for i, j in combs:
            vs[i, j] = compare_tukey(i, j)
    else:
        for i, j in combs:
            vs[i, j] = compare_stats(i, j)

        if p_adjust is not None:
            vs[tri_upper] = multipletests(vs[tri_upper], method=p_adjust)[1]

    vs[tri_lower] = np.transpose(vs)[tri_lower]
    np.fill_diagonal(vs, 1)
    return DataFrame(vs, index=groups, columns=groups)


def posthoc_npm_test(
    a: Union[list, np.ndarray, DataFrame],
    val_col: Optional[str] = None,
    group_col: Optional[str] = None,
    alternative: Literal["greater", "less"] = "greater",
    nperm: int = 1000,
    sort: bool = False,
) -> DataFrame:
    """Calculate pairwise comparisons using Nashimoto and Wright´s all-pairs
    comparison procedure (NPM test) for simply ordered mean ranksums.

    NPM test is basically an extension of Nemenyi´s procedure for testing
    increasingly ordered alternatives [1]_.

    Parameters
    ----------
    a : array_like or pandas DataFrame object
        An array, any object exposing the array interface or a pandas
        DataFrame.

    val_col : str, optional
        Name of a DataFrame column that contains dependent variable values
        (test or response variable). Values should have a non-nominal scale.
        Must be specified if `a` is a pandas DataFrame object.

    group_col : str, optional
        Name of a DataFrame column that contains independent variable values
        (grouping or predictor variable). Values should have a nominal scale
        (categorical). Must be specified if `a` is a pandas DataFrame object.

    sort : bool, optional
        If True, sort data by block and group columns.

    alternative : str, optional
        Alternative hypothesis being tested.
        Can be either "greater" (by default) or "less".

    nperm : int, optional
        Number of permutations to perform for calculating p values.

    Returns
    -------
    result : pandas.DataFrame
        P values.

    Notes
    -----
    An asymetric permutation test is conducted for calculating p values.

    References
    ----------
    .. [1] Nashimoto, K., Wright, F.T., (2005), Multiple comparison procedures for
        detecting differences in simply ordered means. Comput. Statist. Data
        Anal. 48, 291--306.

    Examples
    --------
    >>> x = np.array([[102,109,114,120,124],
                      [110,112,123,130,145],
                      [132,141,156,160,172]])
    >>> sp.posthoc_npm_test(x)
    """
    x, _val_col, _group_col = __convert_to_df(a, val_col, group_col)
    x = x.sort_values(by=[_group_col], ascending=True) if sort else x
    if alternative == "less":
        x[_val_col] *= -1

    groups = x[_group_col].unique()
    k = groups.size
    n = x.shape[0]
    sigma = np.sqrt(n * (n + 1) / 12.0)

    def compare(x, ix):
        x0 = x.copy()
        x0.loc[:, _val_col] = x0.loc[ix, _val_col].values
        x0["ranks"] = x0[_val_col].rank()
        ri = x0.groupby(_group_col, observed=True)["ranks"].mean()
        ni = x0.groupby(_group_col, observed=True)[_val_col].count()
        is_balanced = all(ni == n)
        stat = np.ones((k, k))

        for i in range(k - 1):
            for j in range(i + 1, k):
                m = np.arange(i, j)
                if is_balanced:
                    tmp = [
                        (ri.loc[groups[j]] - ri.loc[groups[_mi]]) / (sigma / np.sqrt(n))
                        for _mi in m
                    ]
                else:
                    tmp = [
                        (ri.loc[groups[j]] - ri.loc[groups[_mi]])
                        / (sigma * np.sqrt(1.0 / ni.loc[groups[_mi]] + 1.0 / ni.loc[groups[j]]))
                        for _mi in m
                    ]
                stat[j, i] = np.max(tmp)
        return stat

    stat = compare(x, x.index)

    mt = np.zeros((nperm, k, k))
    for i in range(nperm):
        ix = np.random.permutation(x.index)
        mt[i] = compare(x, ix)

    p_values = (mt >= stat).sum(axis=0) / nperm

    tri_upper = np.triu_indices(p_values.shape[0], 1)
    p_values[tri_upper] = np.transpose(p_values)[tri_upper]
    np.fill_diagonal(p_values, 1)

    return DataFrame(p_values, index=groups, columns=groups)


def posthoc_siegel_friedman(
    a: Union[list, np.ndarray, DataFrame],
    y_col: Optional[str] = None,
    group_col: Optional[str] = None,
    block_col: Optional[str] = None,
    block_id_col: Optional[str] = None,
    p_adjust: Optional[str] = None,
    melted: bool = False,
    sort: bool = False,
) -> DataFrame:
    """Siegel and Castellan´s All-Pairs Comparisons Test for Unreplicated Blocked
    Data. See authors' paper for additional information [1]_.

    Parameters
    ----------
    a : array_like or pandas DataFrame object
        An array, any object exposing the array interface or a pandas
        DataFrame.

        If `melted` is set to False (default), `a` is a typical matrix of
        block design, i.e. rows are blocks, and columns are groups. In this
        case you do not need to specify col arguments.

        If `a` is an array and `melted` is set to True,
        y_col, block_col and group_col must specify the indices of columns
        containing elements of correspondary type.

        If `a` is a Pandas DataFrame and `melted` is set to True,
        y_col, block_col and group_col must specify columns names (strings).

    y_col : str or int
        Must be specified if `a` is a melted pandas DataFrame object.
        Name of the column that contains y data.

    block_col : str or int
        Must be specified if `a` is a melted pandas DataFrame object.
        Name of the column that contains blocking factor values.

    group_col : str or int
        Must be specified if `a` is a melted pandas DataFrame object.
        Name of the column that contains treatment (group) factor values.

    block_id_col : str or int
        Must be specified if `a` is a melted pandas DataFrame object.
        Name of the column that contains identifiers of blocking factor values.

    melted : bool, optional
        Specifies if data are given as melted columns "y", "blocks", and
        "groups".

    sort : bool, optional
        If True, sort data by block and group columns.

    p_adjust : str, optional
        Method for adjusting p values. See statsmodels.sandbox.stats.multicomp for details. Available methods are:
        'bonferroni' : one-step correction
        'sidak' : one-step correction
        'holm-sidak' : step-down method using Sidak adjustments
        'holm' : step-down method using Bonferroni adjustments
        'simes-hochberg' : step-up method  (independent)
        'hommel' : closed method based on Simes tests (non-negative)
        'fdr_bh' : Benjamini/Hochberg  (non-negative)
        'fdr_by' : Benjamini/Yekutieli (negative)
        'fdr_tsbh' : two stage fdr correction (non-negative)
        'fdr_tsbky' : two stage fdr correction (non-negative)

    Returns
    -------
    result : pandas.DataFrame
        P values.

    Notes
    -----
    For all-pairs comparisons in a two factorial unreplicated complete block design
    with non-normally distributed residuals, Siegel and Castellan's test can be
    performed on Friedman-type ranked data.

    References
    ----------
    .. [1] S. Siegel, N. J. Castellan Jr. (1988), Nonparametric Statistics for the
        Behavioral Sciences. 2nd ed. New York: McGraw-Hill.

    Examples
    --------
    >>> x = np.array([[31,27,24],[31,28,31],[45,29,46],[21,18,48],[42,36,46],[32,17,40]])
    >>> sp.posthoc_siegel_friedman(x)
    """

    def compare_stats(i, j):
        dif = np.abs(R[groups[i]] - R[groups[j]])
        zval = dif / np.sqrt(k * (k + 1.0) / (6.0 * n))
        return zval

    x, _y_col, _group_col, _block_col, _block_id_col = __convert_to_block_df(
        a, y_col, group_col, block_col, block_id_col, melted
    )
    x = x.sort_values(by=[_group_col, _block_col], ascending=True) if sort else x
    x.dropna(inplace=True)

    groups = x[_group_col].unique()
    k = groups.size
    n = x[_block_id_col].unique().size

    x["mat"] = x.groupby(_block_id_col, observed=True)[_y_col].rank()
    R = x.groupby(_group_col, observed=True)["mat"].mean()

    vs = np.zeros((k, k), dtype=float)
    combs = it.combinations(range(k), 2)

    tri_upper = np.triu_indices(vs.shape[0], 1)
    tri_lower = np.tril_indices(vs.shape[0], -1)
    vs[:, :] = 0

    for i, j in combs:
        vs[i, j] = compare_stats(i, j)
    vs = 2.0 * ss.norm.sf(np.abs(vs))
    vs[vs > 1] = 1.0

    if p_adjust:
        vs[tri_upper] = multipletests(vs[tri_upper], method=p_adjust)[1]

    vs[tri_lower] = np.transpose(vs)[tri_lower]
    np.fill_diagonal(vs, 1)
    return DataFrame(vs, index=groups, columns=groups)


def posthoc_miller_friedman(
    a: Union[list, np.ndarray, DataFrame],
    y_col: Optional[str] = None,
    group_col: Optional[str] = None,
    block_col: Optional[str] = None,
    block_id_col: Optional[str] = None,
    melted: bool = False,
    sort: bool = False,
) -> DataFrame:
    """Miller´s All-Pairs Comparisons Test for Unreplicated Blocked Data.
    The p-values are computed from the chi-square distribution [1]_, [2]_,
    [3]_.

    Parameters
    ----------
    a : array_like or pandas DataFrame object
        An array, any object exposing the array interface or a pandas
        DataFrame.

        If `melted` is set to False (default), `a` is a typical matrix of
        block design, i.e. rows are blocks, and columns are groups. In this
        case you do not need to specify col arguments.

        If `a` is an array and `melted` is set to True,
        y_col, block_col and group_col must specify the indices of columns
        containing elements of correspondary type.

        If `a` is a Pandas DataFrame and `melted` is set to True,
        y_col, block_col and group_col must specify columns names (strings).

    y_col : str or int
        Must be specified if `a` is a pandas DataFrame object.
        Name of the column that contains y data.

    group_col : str or int
        Must be specified if `a` is a pandas DataFrame object.
        Name of the column that contains treatment (group) factor values.

    block_col : str or int
        Must be specified if `a` is a pandas DataFrame object.
        Name of the column that contains blocking factor values.

    block_id_col : str or int
        Must be specified if `a` is a melted pandas DataFrame object.
        Name of the column that contains identifiers of blocking factor values.

    melted : bool, optional
        Specifies if data are given as melted columns "y", "blocks", and
        "groups".

    sort : bool, optional
        If True, sort data by block and group columns.

    Returns
    -------
    result : pandas.DataFrame
        P values.

    Notes
    -----
    For all-pairs comparisons in a two factorial unreplicated complete block
    design with non-normally distributed residuals, Miller's test can be
    performed on Friedman-type ranked data.

    References
    ----------
    .. [1] J. Bortz J, G. A. Lienert, K. Boehnke (1990), Verteilungsfreie
        Methoden in der Biostatistik. Berlin: Springerself.

    .. [2] R. G. Miller Jr. (1996), Simultaneous statistical inference. New
        York: McGraw-Hill.

    .. [3] E. L. Wike (2006), Data Analysis. A Statistical Primer for Psychology
        Students. New Brunswick: Aldine Transaction.

    Examples
    --------
    >>> x = np.array([[31,27,24],[31,28,31],[45,29,46],[21,18,48],[42,36,46],[32,17,40]])
    >>> sp.posthoc_miller_friedman(x)
    """

    def compare_stats(i, j):
        dif = np.abs(R[groups[i]] - R[groups[j]])
        qval = dif / np.sqrt(k * (k + 1.0) / (6.0 * n))
        return qval

    x, _y_col, _group_col, _block_col, _block_id_col = __convert_to_block_df(
        a, y_col, group_col, block_col, block_id_col, melted
    )
    x = x.sort_values(by=[_group_col, _block_col], ascending=True) if sort else x
    x.dropna(inplace=True)

    groups = x[_group_col].unique()
    k = groups.size
    n = x[_block_id_col].unique().size

    x["mat"] = x.groupby(_block_id_col, observed=True)[_y_col].rank()
    R = x.groupby(_group_col, observed=True)["mat"].mean()

    vs = np.zeros((k, k), dtype=float)
    combs = it.combinations(range(k), 2)

    tri_lower = np.tril_indices(vs.shape[0], -1)
    vs[:, :] = 0

    for i, j in combs:
        vs[i, j] = compare_stats(i, j)
    vs = vs**2
    vs = ss.chi2.sf(vs, k - 1)

    vs[tri_lower] = np.transpose(vs)[tri_lower]
    np.fill_diagonal(vs, 1)
    return DataFrame(vs, index=groups, columns=groups)


def posthoc_durbin(
    a: Union[list, np.ndarray, DataFrame],
    y_col: Optional[str] = None,
    group_col: Optional[str] = None,
    block_col: Optional[str] = None,
    block_id_col: Optional[str] = None,
    p_adjust: Optional[str] = None,
    melted: bool = False,
    sort: bool = False,
) -> DataFrame:
    """Pairwise post hoc test for multiple comparisons of rank sums according to
    Durbin and Conover for a two-way balanced incomplete block design (BIBD). See
    references for additional information [1]_, [2]_.

    Parameters
    ----------
    a : array_like or pandas DataFrame object
        An array, any object exposing the array interface or a pandas
        DataFrame.

        If `melted` is set to False (default), `a` is a typical matrix of block design,
        i.e. rows are blocks, and columns are groups. In this case you do
        not need to specify col arguments.

        If `a` is an array and `melted` is set to True,
        y_col, block_col and group_col must specify the indices of columns
        containing elements of correspondary type.

        If `a` is a Pandas DataFrame and `melted` is set to True,
        y_col, block_col and group_col must specify columns names (string).

    y_col : str or int
        Must be specified if `a` is a pandas DataFrame object.
        Name of the column that contains y data.

    group_col : str or int
        Must be specified if `a` is a pandas DataFrame object.
        Name of the column that contains treatment (group) factor values.

    block_col : str or int
        Must be specified if `a` is a pandas DataFrame object.
        Name of the column that contains blocking factor values.

    block_id_col : str or int
        Must be specified if `a` is a pandas DataFrame object.
        Name of the column that contains identifiers of blocking factor values.

    melted : bool, optional
        Specifies if data are given as melted columns "y", "blocks", and
        "groups".

    sort : bool, optional
        If True, sort data by block and group columns.

    p_adjust : str, optional
        Method for adjusting p values. See statsmodels.sandbox.stats.multicomp
        for details. Available methods are:
        'bonferroni' : one-step correction
        'sidak' : one-step correction
        'holm-sidak' : step-down method using Sidak adjustments
        'holm' : step-down method using Bonferroni adjustments
        'simes-hochberg' : step-up method  (independent)
        'hommel' : closed method based on Simes tests (non-negative)
        'fdr_bh' : Benjamini/Hochberg  (non-negative)
        'fdr_by' : Benjamini/Yekutieli (negative)
        'fdr_tsbh' : two stage fdr correction (non-negative)
        'fdr_tsbky' : two stage fdr correction (non-negative)

    Returns
    -------
    result : pandas.DataFrame
        P values.

    References
    ----------
    .. [1] W. J. Conover and R. L. Iman (1979), On multiple-comparisons procedures,
        Tech. Rep. LA-7677-MS, Los Alamos Scientific Laboratory.
    .. [2] W. J. Conover (1999), Practical nonparametric Statistics,
        3rd. edition, Wiley.

    Examples
    --------
    >>> x = np.array([[31,27,24],[31,28,31],[45,29,46],[21,18,48],[42,36,46],[32,17,40]])
    >>> sp.posthoc_durbin(x)
    """
    x, _y_col, _group_col, _block_col, _block_id_col = __convert_to_block_df(
        a, y_col, group_col, block_col, block_id_col, melted
    )

    def compare_stats(i, j):
        dif = np.abs(rj[groups[i]] - rj[groups[j]])
        tval = dif / denom
        pval = 2.0 * ss.t.sf(np.abs(tval), df=df)
        return pval

    x = x.sort_values(by=[_block_col, _group_col], ascending=True) if sort else x
    x.dropna(inplace=True)

    groups = x[_group_col].unique()
    t = len(groups)
    b = x[_block_id_col].unique().size
    r = b
    k = t
    x["y_ranked"] = x.groupby(_block_id_col, observed=True)[_y_col].rank()
    rj = x.groupby(_group_col, observed=True)["y_ranked"].sum()
    A = (x["y_ranked"] ** 2).sum()
    C = (b * k * (k + 1) ** 2) / 4.0
    D = (rj.to_numpy() ** 2).sum() - r * C
    T1 = (t - 1) / (A - C) * D
    denom = np.sqrt(((A - C) * 2 * r) / (b * k - b - t + 1) * (1 - T1 / (b * (k - 1))))
    df = b * k - b - t + 1
    print(t, b, r, k)
    print(A, C, D, T1)
    print(denom, df)

    vs = np.zeros((t, t), dtype=float)
    combs = it.combinations(range(t), 2)

    tri_upper = np.triu_indices(vs.shape[0], 1)
    tri_lower = np.tril_indices(vs.shape[0], -1)
    vs[:, :] = 0

    for i, j in combs:
        vs[i, j] = compare_stats(i, j)

    if p_adjust:
        vs[tri_upper] = multipletests(vs[tri_upper], method=p_adjust)[1]

    vs[tri_lower] = np.transpose(vs)[tri_lower]
    np.fill_diagonal(vs, 1)
    return DataFrame(vs, index=groups, columns=groups)


def posthoc_anderson(
    a: Union[list, np.ndarray, DataFrame],
    val_col: Optional[str] = None,
    group_col: Optional[str] = None,
    midrank: bool = True,
    p_adjust: Optional[str] = None,
    sort: bool = False,
) -> DataFrame:
    """Anderson-Darling Pairwise Test for k-samples. Tests the null hypothesis
    that k-samples are drawn from the same population without having to specify
    the distribution function of that population [1]_.

    Parameters
    ----------
    a : array_like or pandas DataFrame object
        An array, any object exposing the array interface or a pandas
        DataFrame.

    val_col : str, optional
        Name of a DataFrame column that contains dependent variable values (test
        or response variable). Values should have a non-nominal scale. Must be
        specified if `a` is a pandas DataFrame object.

    group_col : str, optional
        Name of a DataFrame column that contains independent variable values
        (grouping or predictor variable). Values should have a nominal scale
        (categorical). Must be specified if `a` is a pandas DataFrame object.

    midrank : bool, optional
        Type of Anderson-Darling test which is computed. If set to True (default), the
        midrank test applicable to continuous and discrete populations is performed. If
        False, the right side empirical distribution is used.

    sort : bool, optional
        If True, sort data by block and group columns.

    p_adjust : str, optional
        Method for adjusting p values. See statsmodels.sandbox.stats.multicomp for details. Available methods are:
        'bonferroni' : one-step correction
        'sidak' : one-step correction
        'holm-sidak' : step-down method using Sidak adjustments
        'holm' : step-down method using Bonferroni adjustments
        'simes-hochberg' : step-up method  (independent)
        'hommel' : closed method based on Simes tests (non-negative)
        'fdr_bh' : Benjamini/Hochberg  (non-negative)
        'fdr_by' : Benjamini/Yekutieli (negative)
        'fdr_tsbh' : two stage fdr correction (non-negative)
        'fdr_tsbky' : two stage fdr correction (non-negative)

    Returns
    -------
    result : pandas.DataFrame
        P values.

    References
    ----------
    .. [1] F.W. Scholz, M.A. Stephens (1987), K-Sample Anderson-Darling Tests,
        Journal of the American Statistical Association, Vol. 82, pp. 918-924.

    Examples
    --------
    >>> x = np.array([[2.9, 3.0, 2.5, 2.6, 3.2], [3.8, 2.7, 4.0, 2.4], [2.8, 3.4, 3.7, 2.2, 2.0]])
    >>> sp.posthoc_anderson(x)
    """
    x, _val_col, _group_col = __convert_to_df(a, val_col, group_col)
    x = x.sort_values(by=[_group_col], ascending=True) if sort else x

    groups = x[_group_col].unique()
    k = groups.size
    vs = np.zeros((k, k), dtype=float)
    combs = it.combinations(range(k), 2)

    tri_upper = np.triu_indices(vs.shape[0], 1)
    tri_lower = np.tril_indices(vs.shape[0], -1)
    vs[:, :] = 0

    for i, j in combs:
        vs[i, j] = ss.anderson_ksamp(
            [
                x.loc[x[_group_col] == groups[i], _val_col],
                x.loc[x[_group_col] == groups[j], _val_col],
            ],
            midrank=midrank,
            method=ss.PermutationMethod(),
        )[2]

    if p_adjust:
        vs[tri_upper] = multipletests(vs[tri_upper], method=p_adjust)[1]

    vs[tri_lower] = np.transpose(vs)[tri_lower]
    np.fill_diagonal(vs, 1)
    return DataFrame(vs, index=groups, columns=groups)


def posthoc_quade(
    a: Union[list, np.ndarray, DataFrame],
    y_col: Optional[str] = None,
    group_col: Optional[str] = None,
    block_col: Optional[str] = None,
    block_id_col: Optional[str] = None,
    dist: str = "t",
    p_adjust: Optional[str] = None,
    melted: bool = False,
    sort: bool = False,
) -> DataFrame:
    """Calculate pairwise comparisons using Quade's post hoc test for
    unreplicated blocked data. This test is usually conducted if significant
    results were obtained by the omnibus test [1]_, [2]_, [3]_.

    Parameters
    ----------
    a : array_like or pandas DataFrame object
        An array, any object exposing the array interface or a pandas
        DataFrame.

        If `melted` is set to False (default), `a` is a typical matrix of
        block design, i.e. rows are blocks, and columns are groups. In this
        case you do not need to specify col arguments.

        If `a` is an array and `melted` is set to True,
        y_col, block_col and group_col must specify the indices of columns
        containing elements of correspondary type.

        If `a` is a Pandas DataFrame and `melted` is set to True,
        y_col, block_col and group_col must specify columns names (string).

    y_col : str or int, optional
        Must be specified if `a` is a pandas DataFrame object.
        Name of the column that contains y data.

    block_col : str or int
        Must be specified if `a` is a pandas DataFrame object.
        Name of the column that contains blocking factor values.

    group_col : str or int
        Must be specified if `a` is a pandas DataFrame object.
        Name of the column that contains treatment (group) factor values.

    dist : str, optional
        Method for determining p values.
        The default distribution is "t", else "normal".

    melted : bool, optional
        Specifies if data are given as melted columns "y", "blocks", and
        "groups".

    sort : bool, optional
        If True, sort data by block and group columns.

    p_adjust : str, optional
        Method for adjusting p values.
        See statsmodels.sandbox.stats.multicomp for details.
        Available methods are:
        'bonferroni' : one-step correction
        'sidak' : one-step correction
        'holm-sidak' : step-down method using Sidak adjustments
        'holm' : step-down method using Bonferroni adjustments
        'simes-hochberg' : step-up method  (independent)
        'hommel' : closed method based on Simes tests (non-negative)
        'fdr_bh' : Benjamini/Hochberg  (non-negative)
        'fdr_by' : Benjamini/Yekutieli (negative)
        'fdr_tsbh' : two stage fdr correction (non-negative)
        'fdr_tsbky' : two stage fdr correction (non-negative)

    Returns
    -------
    result : pandas.DataFrame
        P values.

    References
    ----------
    .. [1] W. J. Conover (1999), Practical nonparametric Statistics, 3rd. Edition,
        Wiley.

    .. [2] N. A. Heckert and J. J. Filliben (2003). NIST Handbook 148: Dataplot
        Reference Manual, Volume 2: Let Subcommands and Library Functions.
        National Institute of Standards and Technology Handbook Series, June 2003.

    .. [3] D. Quade (1979), Using weighted rankings in the analysis of complete
        blocks with additive block effects. Journal of the American Statistical
        Association, 74, 680-683.

    Examples
    --------
    >>> x = np.array([[31,27,24],[31,28,31],[45,29,46],[21,18,48],[42,36,46],[32,17,40]])
    >>> sp.posthoc_quade(x)
    """

    def compare_stats_t(i, j):
        dif = np.abs(S[groups[i]] - S[groups[j]])
        tval = dif / denom
        pval = 2.0 * ss.t.sf(np.abs(tval), df=(b - 1) * (k - 1))
        return pval

    def compare_stats_norm(i, j):
        dif = np.abs(W[groups[i]] * ff - W[groups[j]] * ff)
        zval = dif / denom
        pval = 2.0 * ss.norm.sf(np.abs(zval))
        return pval

    x, _y_col, _group_col, _block_col, _block_id_col = __convert_to_block_df(
        a, y_col, group_col, block_col, block_id_col, melted
    )

    x = x.sort_values(by=[_block_col, _group_col], ascending=True) if sort else x
    x.dropna(inplace=True)

    groups = x[_group_col].unique()
    k = len(groups)
    b = x[_block_id_col].unique().size

    x["r"] = x.groupby(_block_id_col, observed=True)[_y_col].rank()
    q = Series(
        x.groupby(_block_id_col, observed=True)[_y_col].max()
        - x.groupby(_block_id_col, observed=True)[_y_col].min().to_numpy()
    ).rank()
    x["rr"] = x["r"] - (k + 1) / 2
    x["s"] = x.apply(lambda row: row["rr"] * q[row[_block_id_col]], axis=1)
    x["w"] = x.apply(lambda row: row["r"] * q[row[_block_id_col]], axis=1)

    A = (x["s"] ** 2).sum()
    S = x.groupby(_group_col, observed=True)["s"].sum()
    B = np.sum(S.to_numpy() ** 2) / b
    W = x.groupby(_group_col, observed=True)["w"].sum()

    vs = np.zeros((k, k), dtype=float)
    combs = it.combinations(range(k), 2)

    tri_upper = np.triu_indices(vs.shape[0], 1)
    tri_lower = np.tril_indices(vs.shape[0], -1)
    vs[:, :] = 0

    if dist == "t":
        denom = np.sqrt((2 * b * (A - B)) / ((b - 1) * (k - 1)))

        for i, j in combs:
            vs[i, j] = compare_stats_t(i, j)

    else:
        n = b * k
        denom = np.sqrt((k * (k + 1.0) * (2.0 * n + 1.0) * (k - 1.0)) / (18.0 * n * (n + 1.0)))
        ff = 1.0 / (b * (b + 1.0) / 2.0)

        for i, j in combs:
            vs[i, j] = compare_stats_norm(i, j)

    if p_adjust:
        vs[tri_upper] = multipletests(vs[tri_upper], method=p_adjust)[1]

    vs[tri_lower] = np.transpose(vs)[tri_lower]
    np.fill_diagonal(vs, 1)
    return DataFrame(vs, index=groups, columns=groups)


def posthoc_vanwaerden(
    a: Union[list, np.ndarray, DataFrame],
    val_col: Optional[str] = None,
    group_col: Optional[str] = None,
    sort: bool = False,
    p_adjust: Optional[str] = None,
) -> DataFrame:
    """Van der Waerden's test for pairwise multiple comparisons between group
    levels. See references for additional information [1]_, [2]_.

    Parameters
    ----------
    a : array_like or pandas DataFrame object
        An array, any object exposing the array interface or a pandas
        DataFrame.

    val_col : str, optional
        Name of a DataFrame column that contains dependent variable values (test
        or response variable). Values should have a non-nominal scale. Must be
        specified if `a` is a pandas DataFrame object.

    group_col : str, optional
        Name of a DataFrame column that contains independent variable values
        (grouping or predictor variable). Values should have a nominal scale
        (categorical). Must be specified if `a` is a pandas DataFrame object.

    sort : bool, optional
        If True, sort data by block and group columns.

    p_adjust : str, optional
        Method for adjusting p values.
        See statsmodels.sandbox.stats.multicomp for details.
        Available methods are:
        'bonferroni' : one-step correction
        'sidak' : one-step correction
        'holm-sidak' : step-down method using Sidak adjustments
        'holm' : step-down method using Bonferroni adjustments
        'simes-hochberg' : step-up method  (independent)
        'hommel' : closed method based on Simes tests (non-negative)
        'fdr_bh' : Benjamini/Hochberg  (non-negative)
        'fdr_by' : Benjamini/Yekutieli (negative)
        'fdr_tsbh' : two stage fdr correction (non-negative)
        'fdr_tsbky' : two stage fdr correction (non-negative)

    Returns
    -------
    result : pandas.DataFrame
        P values.

    Notes
    -----
    For one-factorial designs with samples that do not meet the assumptions for
    one-way-ANOVA and subsequent post hoc tests, the van der Waerden test using
    normal scores can be employed. Provided that significant differences were
    detected by this global test, one may be interested in applying post hoc
    tests according to van der Waerden for pairwise multiple comparisons of the
    group levels.

    There is no tie correction applied in this function.

    References
    ----------
    .. [1] W. J. Conover and R. L. Iman (1979), On multiple-comparisons procedures,
        Tech. Rep. LA-7677-MS, Los Alamos Scientific Laboratory.

    .. [2] B. L. van der Waerden (1952) Order tests for the two-sample problem and
        their power, Indagationes Mathematicae, 14, 453-458.

    Examples
    --------
    >>> x = np.array([[10,'a'], [59,'a'], [76,'b'], [10, 'b']])
    >>> sp.posthoc_vanwaerden(x, val_col = 0, group_col = 1)
    """
    x, _val_col, _group_col = __convert_to_df(a, val_col, group_col)
    x = x.sort_values(by=[_group_col], ascending=True) if sort else x

    groups = x[_group_col].unique()
    n = x[_val_col].size
    k = groups.size
    r = ss.rankdata(x[_val_col])
    x["z_scores"] = ss.norm.ppf(r / (n + 1))

    aj = x.groupby(_group_col, observed=True)["z_scores"].sum().to_numpy()
    nj = x.groupby(_group_col, observed=True)["z_scores"].count()
    s2 = (1.0 / (n - 1.0)) * (x["z_scores"] ** 2.0).sum()
    sts = (1.0 / s2) * np.sum(aj**2.0 / nj)
    A = aj / nj

    vs = np.zeros((k, k), dtype=float)
    combs = it.combinations(range(k), 2)

    tri_upper = np.triu_indices(vs.shape[0], 1)
    tri_lower = np.tril_indices(vs.shape[0], -1)
    vs[:, :] = 0

    def compare_stats(i, j):
        dif = np.abs(A[groups[i]] - A[groups[j]])
        B = 1.0 / nj[groups[i]] + 1.0 / nj[groups[j]]
        tval = dif / np.sqrt(s2 * (n - 1.0 - sts) / (n - k) * B)
        pval = 2.0 * ss.t.sf(np.abs(tval), df=n - k)
        return pval

    for i, j in combs:
        vs[i, j] = compare_stats(i, j)

    if p_adjust:
        vs[tri_upper] = multipletests(vs[tri_upper], method=p_adjust)[1]

    vs[tri_lower] = np.transpose(vs)[tri_lower]
    np.fill_diagonal(vs, 1)
    return DataFrame(vs, index=groups, columns=groups)


def posthoc_dunnett(
    a: Union[list, np.ndarray, DataFrame],
    val_col: Optional[str] = None,
    group_col: Optional[str] = None,
    control: Optional[str] = None,
    alternative: Literal["two-sided", "less", "greater"] = "two-sided",
    sort: bool = False,
    to_matrix: bool = True,
) -> Union[Series, DataFrame]:
    """
    Dunnett's test [1, 2, 3] for multiple comparisons against a control group, used after parametric
    ANOVA. The control group is specified by the `control` parameter.

    Parameters
    ----------
    a : array_like or pandas DataFrame object
        An array, any object exposing the array interface or a pandas
        DataFrame. Array must be two-dimensional.

    val_col : str, optional
        Name of a DataFrame column that contains dependent variable values (test
        or response variable). Values should have a non-nominal scale. Must be
        specified if `a` is a pandas DataFrame object.

    group_col : str, optional
        Name of a DataFrame column that contains independent variable values
        (grouping or predictor variable). Values should have a nominal scale
        (categorical). Must be specified if `a` is a pandas DataFrame object.

    control : str, optional
        Name of the control group within the `group_col` column. Values should
        have a nominal scale (categorical). Must be specified if `a` is a pandas
        DataFrame.

    alternative : ['two-sided', 'less', or 'greater'], optional
        Whether to get the p-value for the one-sided hypothesis
        ('less' or 'greater') or for the two-sided hypothesis ('two-sided').
        Defaults to 'two-sided'.

    sort : bool, optional
        Specifies whether to sort DataFrame by group_col or not. Recommended
        unless you sort your data manually.

    to_matrix: bool, optional
        Specifies whether to return a DataFrame or a Series. If True, a DataFrame
        is returned with some NaN values since it's not pairwise comparison.
        Default is True.

    Returns
    -------
    result : pandas.Series or pandas.DataFrame
        P values.

    References
    ----------
    .. [1] Charles W. Dunnett (1955). “A Multiple Comparison Procedure for Comparing Several Treatments with a Control.”
    .. [2] https://en.wikipedia.org/wiki/Dunnett%27s_test
    .. [3] https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.dunnett.html#id1
    """
    x, _val_col, _group_col = __convert_to_df(a, val_col, group_col)
    x = x.sort_values(by=[_group_col], ascending=True) if sort else x
    x = x.set_index(_group_col)[_val_col]
    x_embedded = x.groupby(_group_col, observed=True).agg(lambda y: y.dropna().tolist())
    control_data = x_embedded.loc[control]
    treatment_data = x_embedded.drop(control)

    pvals = ss.dunnett(*treatment_data, control=control_data, alternative=alternative).pvalue

    multi_index = MultiIndex.from_product([[control], treatment_data.index.tolist()])
    dunnett_sr = Series(pvals, index=multi_index)

    if not to_matrix:
        return dunnett_sr

    else:
        levels = x.index.unique().to_numpy()
        result_df = DataFrame(index=levels, columns=levels)

        for pair in dunnett_sr.index:
            a, b = pair
            result_df.loc[a, b] = dunnett_sr[pair]
            result_df.loc[b, a] = dunnett_sr[pair]
        result_df.loc[control, control] = 1.0
        return result_df


def posthoc_ttest(
    a: Union[list, np.ndarray, DataFrame],
    val_col: Optional[str] = None,
    group_col: Optional[str] = None,
    pool_sd: bool = False,
    equal_var: bool = True,
    p_adjust: Optional[str] = None,
    sort: bool = False,
) -> DataFrame:
    """Pairwise T test for multiple comparisons of independent groups. May be
    used after a parametric ANOVA to do pairwise comparisons.

    Parameters
    ----------
    a : array_like or pandas DataFrame object
        An array, any object exposing the array interface or a pandas
        DataFrame. Array must be two-dimensional.

    val_col : str, optional
        Name of a DataFrame column that contains dependent variable values (test
        or response variable). Values should have a non-nominal scale. Must be
        specified if `a` is a pandas DataFrame object.

    group_col : str, optional
        Name of a DataFrame column that contains independent variable values
        (grouping or predictor variable). Values should have a nominal scale
        (categorical). Must be specified if `a` is a pandas DataFrame object.

    equal_var : bool, optional
        If True (default), perform a standard independent test
        that assumes equal population variances [1]_.
        If False, perform Welch's t-test, which does not assume equal
        population variance [2]_.

    pool_sd : bool, optional
        Calculate a common SD for all groups and use that for all
        comparisons (this can be useful if some groups are small).
        This method does not actually call scipy ttest_ind() function,
        so extra arguments are ignored. Default is False.

    p_adjust : str, optional
        Method for adjusting p values.
        See statsmodels.sandbox.stats.multicomp for details.
        Available methods are:
        'bonferroni' : one-step correction
        'sidak' : one-step correction
        'holm-sidak' : step-down method using Sidak adjustments
        'holm' : step-down method using Bonferroni adjustments
        'simes-hochberg' : step-up method  (independent)
        'hommel' : closed method based on Simes tests (non-negative)
        'fdr_bh' : Benjamini/Hochberg  (non-negative)
        'fdr_by' : Benjamini/Yekutieli (negative)
        'fdr_tsbh' : two stage fdr correction (non-negative)
        'fdr_tsbky' : two stage fdr correction (non-negative)

    sort : bool, optional
        Specifies whether to sort DataFrame by group_col or not. Recommended
        unless you sort your data manually.

    Returns
    -------
    result : pandas.DataFrame
        P values.

    References
    ----------

    .. [1] http://en.wikipedia.org/wiki/T-test#Independent_two-sample_t-test
    .. [2] http://en.wikipedia.org/wiki/Welch%27s_t_test

    Examples
    --------

    >>> x = [[1,2,3,5,1], [12,31,54, np.nan], [10,12,6,74,11]]
    >>> sp.posthoc_ttest(x, p_adjust = 'holm')
    array([[-1.        ,  0.04600899,  0.31269089],
           [ 0.04600899, -1.        ,  0.6327077 ],
           [ 0.31269089,  0.6327077 , -1.        ]])
    """
    x, _val_col, _group_col = __convert_to_df(a, val_col, group_col)
    x = x.sort_values(by=[_group_col], ascending=True) if sort else x

    groups = x[_group_col].unique()
    k = groups.size
    xg = x.groupby(by=_group_col, observed=True)[_val_col]

    vs = np.zeros((k, k), dtype=float)
    tri_upper = np.triu_indices(vs.shape[0], 1)
    tri_lower = np.tril_indices(vs.shape[0], -1)
    vs[:, :] = 0
    combs = it.combinations(range(k), 2)

    if pool_sd:
        ni = xg.count()
        m = xg.mean()
        sd = xg.std(ddof=1)
        deg_f = ni - 1.0
        total_deg_f = np.sum(deg_f)
        pooled_sd = np.sqrt(np.sum(sd**2.0 * deg_f) / total_deg_f)

        def compare_pooled(i, j):
            diff = m.iloc[i] - m.iloc[j]
            se_diff = pooled_sd * np.sqrt(1.0 / ni.iloc[i] + 1.0 / ni.iloc[j])
            t_value = diff / se_diff
            return 2.0 * ss.t.cdf(-np.abs(t_value), total_deg_f)

        for i, j in combs:
            vs[i, j] = compare_pooled(i, j)
    else:
        for i, j in combs:
            vs[i, j] = ss.ttest_ind(
                xg.get_group(groups[i]), xg.get_group(groups[j]), equal_var=equal_var
            )[1]

    if p_adjust:
        vs[tri_upper] = multipletests(vs[tri_upper], method=p_adjust)[1]

    vs[tri_lower] = np.transpose(vs)[tri_lower]
    np.fill_diagonal(vs, 1)
    return DataFrame(vs, index=groups, columns=groups)


def posthoc_tukey_hsd(
    a: Union[list, np.ndarray, DataFrame],
    val_col: Optional[str] = None,
    group_col: Optional[str] = None,
    sort: bool = True,
) -> DataFrame:
    """Pairwise comparisons with TukeyHSD confidence intervals. This is a
    convenience function to make statsmodels `pairwise_tukeyhsd` method more
    applicable for further use.

    Parameters
    ----------
    x : array_like or pandas Series object, 1d
        An array, any object exposing the array interface, containing dependent
        variable values (test or response variable). Values should have a
        non-nominal scale. NaN values will cause an error (please handle
        manually).

    g : array_like or pandas Series object, 1d
        An array, any object exposing the array interface, containing
        independent variable values (grouping or predictor variable). Values
        should have a nominal scale (categorical).

    alpha : float, optional
        Significance level for the test. Default is 0.05.

    Returns
    -------
    result : pandas.DataFrame
        DataFrame with 0, 1, and -1 values, where 0 is False (not significant),
        1 is True (significant), and -1 is for diagonal elements.

    Examples
    --------
    >>> x = [[1,2,3,4,5], [35,31,75,40,21], [10,6,9,6,1]]
    >>> g = [['a'] * 5, ['b'] * 5, ['c'] * 5]
    >>> sp.posthoc_tukey_hsd(np.concatenate(x), np.concatenate(g))
    """
    x, _val_col, _group_col = __convert_to_df(a, val_col, group_col)
    x = x.sort_values(by=[_group_col, _val_col], ascending=True) if sort else x
    groups = x[_group_col].unique()

    results = ss.tukey_hsd(
        *[
            x.loc[idx, _val_col].to_numpy()
            for idx in x.groupby(_group_col, observed=True).groups.values()
        ]
    )

    return DataFrame(results.pvalue, index=groups, columns=groups)


def posthoc_mannwhitney(
    a: Union[list, np.ndarray, DataFrame],
    val_col: Optional[str] = None,
    group_col: Optional[str] = None,
    use_continuity: bool = True,
    alternative: str = "two-sided",
    p_adjust: Optional[str] = None,
    sort: bool = True,
) -> DataFrame:
    """Pairwise comparisons with Mann-Whitney rank test.

    Parameters
    ----------
    a : array_like or pandas DataFrame object
        An array, any object exposing the array interface or a pandas
        DataFrame. Array must be two-dimensional.

    val_col : str, optional
        Name of a DataFrame column that contains dependent variable values (test
        or response variable). Values should have a non-nominal scale. Must be
        specified if `a` is a pandas DataFrame object.

    group_col : str, optional
        Name of a DataFrame column that contains independent variable values
        (grouping or predictor variable). Values should have a nominal scale
        (categorical). Must be specified if `a` is a pandas DataFrame object.

    use_continuity : bool, optional
        Whether a continuity correction (1/2.) should be taken into account.
        Default is True.

    alternative : ['two-sided', 'less', or 'greater'], optional
        Whether to get the p-value for the one-sided hypothesis
        ('less' or 'greater') or for the two-sided hypothesis ('two-sided').
        Defaults to 'two-sided'.

    p_adjust : str, optional
        Method for adjusting p values.
        See statsmodels.sandbox.stats.multicomp for details.
        Available methods are:
        'bonferroni' : one-step correction
        'sidak' : one-step correction
        'holm-sidak' : step-down method using Sidak adjustments
        'holm' : step-down method using Bonferroni adjustments
        'simes-hochberg' : step-up method  (independent)
        'hommel' : closed method based on Simes tests (non-negative)
        'fdr_bh' : Benjamini/Hochberg  (non-negative)
        'fdr_by' : Benjamini/Yekutieli (negative)
        'fdr_tsbh' : two stage fdr correction (non-negative)
        'fdr_tsbky' : two stage fdr correction (non-negative)

    sort : bool, optional
        Specifies whether to sort DataFrame by group_col or not. Recommended
        unless you sort your data manually.

    Returns
    -------
    result : pandas.DataFrame
        P values.

    Notes
    -----
    Refer to `scipy.stats.mannwhitneyu` reference page for further details.

    Examples
    --------
    >>> x = [[1,2,3,4,5], [35,31,75,40,21], [10,6,9,6,1]]
    >>> sp.posthoc_mannwhitney(x, p_adjust = 'holm')
    """
    x, _val_col, _group_col = __convert_to_df(a, val_col, group_col)
    x = x.sort_values(by=[_group_col, _val_col], ascending=True) if sort else x

    groups = x[_group_col].unique()
    x_len = groups.size
    vs = np.zeros((x_len, x_len))
    xg = x.groupby(_group_col, observed=True)[_val_col]
    tri_upper = np.triu_indices(vs.shape[0], 1)
    tri_lower = np.tril_indices(vs.shape[0], -1)
    vs[:, :] = 0

    combs = it.combinations(range(x_len), 2)

    for i, j in combs:
        vs[i, j] = ss.mannwhitneyu(
            xg.get_group(groups[i]),
            xg.get_group(groups[j]),
            use_continuity=use_continuity,
            alternative=alternative,
        )[1]

    if p_adjust:
        vs[tri_upper] = multipletests(vs[tri_upper], method=p_adjust)[1]

    vs[tri_lower] = np.transpose(vs)[tri_lower]
    np.fill_diagonal(vs, 1)
    return DataFrame(vs, index=groups, columns=groups)


def posthoc_wilcoxon(
    a: Union[list, np.ndarray, DataFrame],
    val_col: Optional[str] = None,
    group_col: Optional[str] = None,
    method: str = "auto",
    zero_method: str = "wilcox",
    correction: bool = False,
    p_adjust: Optional[str] = None,
    sort: bool = False,
) -> DataFrame:
    """Pairwise comparisons with Wilcoxon signed-rank test.

    It is a non-parametric version of the paired T-test for use with
    non-parametric ANOVA.

    Parameters
    ----------
    a : array_like or pandas DataFrame object
        An array, any object exposing the array interface or a pandas
        DataFrame. Array must be two-dimensional.

    val_col : str, optional
        Name of a DataFrame column that contains dependent variable values (test
        or response variable). Values should have a non-nominal scale. Must be
        specified if `a` is a pandas DataFrame object.

    group_col : str, optional
        Name of a DataFrame column that contains independent variable values
        (grouping or predictor variable). Values should have a nominal scale
        (categorical). Must be specified if `a` is a pandas DataFrame object.

    method : {"auto", "exact", "approx"}, optional
        Method to calculate the p-value. Default is "auto".

    zero_method : {"pratt", "wilcox", "zsplit"}, optional
        "pratt": Pratt treatment, includes zero-differences in the ranking
        process (more conservative)
        "wilcox": Wilcox treatment, discards all zero-differences
        "zsplit": Zero rank split, just like Pratt, but spliting the zero rank
        between positive and negative ones

    correction : bool, optional
        If True, apply continuity correction by adjusting the Wilcoxon rank
        statistic by 0.5 towards the mean value when computing the z-statistic.
        Default is False.

    p_adjust : str, optional
        Method for adjusting p values.
        See statsmodels.sandbox.stats.multicomp for details.
        Available methods are:
        'bonferroni' : one-step correction
        'sidak' : one-step correction
        'holm-sidak' : step-down method using Sidak adjustments
        'holm' : step-down method using Bonferroni adjustments
        'simes-hochberg' : step-up method (independent)
        'hommel' : closed method based on Simes tests (non-negative)
        'fdr_bh' : Benjamini/Hochberg (non-negative)
        'fdr_by' : Benjamini/Yekutieli (negative)
        'fdr_tsbh' : two stage fdr correction (non-negative)
        'fdr_tsbky' : two stage fdr correction (non-negative)

    sort : bool, optional
        Specifies whether to sort DataFrame by group_col and val_col or not.
        Default is False.

    Returns
    -------
    result : pandas.DataFrame
        P values.

    Notes
    -----
    Refer to `scipy.stats.wilcoxon` reference page for further details [1]_.

    References
    ----------
    .. [1] https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.wilcoxon.html

    Examples
    --------
    >>> x = [[1,2,3,4,5], [35,31,75,40,21], [10,6,9,6,1]]
    >>> sp.posthoc_wilcoxon(x)
    """
    x, _val_col, _group_col = __convert_to_df(a, val_col, group_col)
    x = x.sort_values(by=[_group_col, _val_col], ascending=True) if sort else x

    groups = x[_group_col].unique()
    x_len = groups.size
    vs = np.zeros((x_len, x_len))
    xg = x.groupby(_group_col, observed=True)[_val_col]
    tri_upper = np.triu_indices(vs.shape[0], 1)
    tri_lower = np.tril_indices(vs.shape[0], -1)
    vs[:, :] = 0

    combs = it.combinations(range(x_len), 2)

    for i, j in combs:
        vs[i, j] = ss.wilcoxon(
            xg.get_group(groups[i]),
            xg.get_group(groups[j]),
            zero_method=zero_method,
            method=method,
            correction=correction,
        )[1]

    if p_adjust:
        vs[tri_upper] = multipletests(vs[tri_upper], method=p_adjust)[1]
    vs[tri_lower] = np.transpose(vs)[tri_lower]
    np.fill_diagonal(vs, 1)
    return DataFrame(vs, index=groups, columns=groups)


def posthoc_scheffe(
    a: Union[list, np.ndarray, DataFrame],
    val_col: Optional[str] = None,
    group_col: Optional[str] = None,
    sort: bool = False,
) -> DataFrame:
    """Scheffe's all-pairs comparisons test for normally distributed data with equal
    group variances. For all-pairs comparisons in an one-factorial layout with
    normally distributed residuals and equal variances Scheffe's test can be
    performed with parametric ANOVA [1]_, [2]_, [3]_.

    A total of m = k(k-1)/2 hypotheses can be tested.

    Parameters
    ----------
    a : Union[list, np.ndarray, DataFrame]
        An array, any object exposing the array interface or a pandas
        DataFrame.

    val_col : str, optional
        Name of a DataFrame column that contains dependent variable values (test
        or response variable). Values should have a non-nominal scale. Must be
        specified if `a` is a pandas DataFrame object.

    group_col : str, optional
        Name of a DataFrame column that contains independent variable values
        (grouping or predictor variable). Values should have a nominal scale
        (categorical). Must be specified if `a` is a pandas DataFrame object.

    sort : bool, optional
        If True, sort data by block and group columns.

    Returns
    -------
    result : pandas.DataFrame
        P values.

    Notes
    -----
    The p values are computed from the F-distribution.

    References
    ----------
    .. [1] J. Bortz (1993) Statistik für Sozialwissenschaftler. 4. Aufl., Berlin:
        Springer.

    .. [2] L. Sachs (1997) Angewandte Statistik, New York: Springer.

    .. [3] H. Scheffe (1953) A Method for Judging all Contrasts in the Analysis
        of Variance. Biometrika 40, 87-110.

    Examples
    --------
    >>> import scikit_posthocs as sp
    >>> import pandas as pd
    >>> x = pd.DataFrame({"a": [1,2,3,5,1], "b": [12,31,54,62,12], "c": [10,12,6,74,11]})
    >>> x = x.melt(var_name='groups', value_name='values')
    >>> sp.posthoc_scheffe(x, val_col='values', group_col='groups')
    """
    x, _val_col, _group_col = __convert_to_df(a, val_col, group_col)
    x = x.sort_values(by=[_group_col], ascending=True) if sort else x

    groups = x[_group_col].unique()
    x_grouped = x.groupby(_group_col, observed=True)[_val_col]
    ni = x_grouped.count()
    xi = x_grouped.mean()
    si = x_grouped.var()
    n = ni.sum()
    sin = 1.0 / (n - groups.size) * np.sum(si * (ni - 1.0))

    def compare(i, j):
        dif = xi.loc[i] - xi.loc[j]
        A = sin * (1.0 / ni.loc[i] + 1.0 / ni.loc[j]) * (groups.size - 1.0)
        f_val = dif**2.0 / A
        return f_val

    vs = np.zeros((groups.size, groups.size), dtype=float)
    tri_lower = np.tril_indices(vs.shape[0], -1)
    vs[:, :] = 0

    combs = it.combinations(range(groups.size), 2)

    for i, j in combs:
        vs[i, j] = compare(groups[i], groups[j])

    vs[tri_lower] = np.transpose(vs)[tri_lower]
    p_values = ss.f.sf(vs, groups.size - 1.0, n - groups.size)

    np.fill_diagonal(p_values, 1)
    return DataFrame(p_values, index=groups, columns=groups)


def posthoc_tamhane(
    a: Union[list, np.ndarray, DataFrame],
    val_col: Optional[str] = None,
    group_col: Optional[str] = None,
    welch: bool = True,
    sort: bool = False,
) -> DataFrame:
    """Tamhane's T2 all-pairs comparison test for normally distributed data with
    unequal variances. Tamhane's T2 test can be performed for all-pairs
    comparisons in an one-factorial layout with normally distributed residuals
    but unequal groups variances. A total of m = k(k-1)/2 hypotheses can be
    tested. The null hypothesis is tested in the two-tailed test against the
    alternative hypothesis [1]_.

    Parameters
    ----------
    a : Union[list, np.ndarray, DataFrame]
        An array, any object exposing the array interface or a pandas
        DataFrame.

    val_col : str, optional
        Name of a DataFrame column that contains dependent variable values (test
        or response variable). Values should have a non-nominal scale. Must be
        specified if `a` is a pandas DataFrame object.

    group_col : str, optional
        Name of a DataFrame column that contains independent variable values
        (grouping or predictor variable). Values should have a nominal scale
        (categorical). Must be specified if `a` is a pandas DataFrame object.

    welch : bool, optional
        If True, use Welch's approximate solution for calculating the degree of
        freedom. T2 test uses the usual df = N - 2 approximation.

    sort : bool, optional
        If True, sort data by block and group columns.

    Returns
    -------
    result : pandas.DataFrame
        P values.

    Notes
    -----
    The p values are computed from the t-distribution and adjusted according to
    Dunn-Sidak.

    References
    ----------
    .. [1] A.C. Tamhane (1979), A Comparison of Procedures for Multiple Comparisons of
        Means with Unequal Variances. Journal of the American Statistical Association,
        74, 471-480.

    Examples
    --------
    >>> import scikit_posthocs as sp
    >>> import pandas as pd
    >>> x = pd.DataFrame({"a": [1,2,3,5,1], "b": [12,31,54,62,12], "c": [10,12,6,74,11]})
    >>> x = x.melt(var_name='groups', value_name='values')
    >>> sp.posthoc_tamhane(x, val_col='values', group_col='groups')
    """
    x, _val_col, _group_col = __convert_to_df(a, val_col, group_col)
    x = x.sort_values(by=[_group_col], ascending=True) if sort else x

    groups = x[_group_col].unique()
    x_grouped = x.groupby(_group_col, observed=True)[_val_col]
    ni = x_grouped.count()
    xi = x_grouped.mean()
    si = x_grouped.var()

    def compare(i, j):
        dif = xi[i] - xi[j]
        A = si[i] / ni[i] + si[j] / ni[j]
        t_val = dif / np.sqrt(A)
        if welch:
            df = A**2.0 / (
                si[i] ** 2.0 / (ni[i] ** 2.0 * (ni[i] - 1.0))
                + si[j] ** 2.0 / (ni[j] ** 2.0 * (ni[j] - 1.0))
            )
        else:
            # checks according to Tamhane (1979, p. 474)
            ok1 = (9.0 / 10.0 <= ni[i] / ni[j]) and (ni[i] / ni[j] <= 10.0 / 9.0)
            ok2 = (9.0 / 10.0 <= (si[i] / ni[i]) / (si[j] / ni[j])) and (
                (si[i] / ni[i]) / (si[j] / ni[j]) <= 10.0 / 9.0
            )
            ok3 = (
                (4.0 / 5.0 <= ni[i] / ni[j])
                and (ni[i] / ni[j] <= 5.0 / 4.0)
                and (1.0 / 2.0 <= (si[i] / ni[i]) / (si[j] / ni[j]))
                and ((si[i] / ni[i]) / (si[j] / ni[j]) <= 2.0)
            )
            ok4 = (
                (2.0 / 3.0 <= ni[i] / ni[j])
                and (ni[i] / ni[j] <= 3.0 / 2.0)
                and (3.0 / 4.0 <= (si[i] / ni[i]) / (si[j] / ni[j]))
                and ((si[i] / ni[i]) / (si[j] / ni[j]) <= 4.0 / 3.0)
            )
            OK = any([ok1, ok2, ok3, ok4])
            if not OK:
                print("Sample sizes or standard errors are not balanced. T2 test is recommended.")
            df = ni[i] + ni[j] - 2.0
        p_val = 2.0 * ss.t.sf(np.abs(t_val), df=df)
        return p_val

    vs = np.zeros((groups.size, groups.size), dtype=float)
    tri_upper = np.triu_indices(vs.shape[0], 1)
    tri_lower = np.tril_indices(vs.shape[0], -1)
    vs[:, :] = 0

    combs = it.combinations(range(groups.size), 2)

    for i, j in combs:
        vs[i, j] = compare(groups[i], groups[j])

    vs[tri_upper] = 1.0 - (1.0 - vs[tri_upper]) ** groups.size
    vs[tri_lower] = np.transpose(vs)[tri_lower]
    vs[vs > 1] = 1

    np.fill_diagonal(vs, 1)
    return DataFrame(vs, index=groups, columns=groups)


def posthoc_tukey(
    a: Union[list, np.ndarray, DataFrame],
    val_col: Optional[str] = None,
    group_col: Optional[str] = None,
    sort: bool = False,
) -> DataFrame:
    """Performs Tukey's all-pairs comparisons test for normally distributed data
    with equal group variances. For all-pairs comparisons in an
    one-factorial layout with normally distributed residuals and equal variances
    Tukey's test can be performed. A total of m = k(k-1)/2 hypotheses can be
    tested. The null hypothesis is tested in the two-tailed test against
    the alternative hypothesis [1]_, [2]_.

    Parameters
    ----------
    a : Union[list, np.ndarray, DataFrame]
        An array, any object exposing the array interface or a pandas
        DataFrame.

    val_col : str, optional
        Name of a DataFrame column that contains dependent variable values (test
        or response variable). Values should have a non-nominal scale. Must be
        specified if `a` is a pandas DataFrame object.

    group_col : str, optional
        Name of a DataFrame column that contains independent variable values
        (grouping or predictor variable). Values should have a nominal scale
        (categorical). Must be specified if `a` is a pandas DataFrame object.

    sort : bool, optional
        If True, sort data by block and group columns.

    Returns
    -------
    result : pandas.DataFrame
        P values.

    Notes
    -----
    The p values are computed from the Tukey-distribution.

    References
    ----------
    .. [1] L. Sachs (1997) Angewandte Statistik, New York: Springer.
    .. [2] J. Tukey (1949) Comparing Individual Means in the Analysis of Variance,
        Biometrics 5, 99-114.

    Examples
    --------
    >>> import scikit_posthocs as sp
    >>> import pandas as pd
    >>> x = pd.DataFrame({"a": [1,2,3,5,1], "b": [12,31,54,62,12], "c": [10,12,6,74,11]})
    >>> x = x.melt(var_name='groups', value_name='values')
    >>> sp.posthoc_tukey(x, val_col='values', group_col='groups')
    """
    x, _val_col, _group_col = __convert_to_df(a, val_col, group_col)
    x = x.sort_values(by=[_group_col], ascending=True) if sort else x
    groups = x[_group_col].unique()
    x_grouped = x.groupby(_group_col, observed=True)[_val_col]

    ni = x_grouped.count()
    n = ni.sum()
    xi = x_grouped.mean()
    si = x_grouped.var()
    sin = 1.0 / (n - groups.size) * np.sum(si * (ni - 1))

    def compare(i, j):
        dif = xi[i] - xi[j]
        A = sin * 0.5 * (1.0 / ni.loc[i] + 1.0 / ni.loc[j])
        q_val = dif / np.sqrt(A)
        return q_val

    vs = np.zeros((groups.size, groups.size), dtype=float)
    tri_upper = np.triu_indices(vs.shape[0], 1)
    tri_lower = np.tril_indices(vs.shape[0], -1)
    vs[:, :] = 0

    combs = it.combinations(range(groups.size), 2)

    for i, j in combs:
        vs[i, j] = compare(groups[i], groups[j])

    vs[tri_upper] = ss.studentized_range.sf(np.abs(vs[tri_upper]), groups.size, n - groups.size)
    vs[tri_lower] = np.transpose(vs)[tri_lower]

    np.fill_diagonal(vs, 1)
    return DataFrame(vs, index=groups, columns=groups)


def posthoc_dscf(
    a: Union[list, np.ndarray, DataFrame],
    val_col: Optional[str] = None,
    group_col: Optional[str] = None,
    sort: bool = False,
) -> DataFrame:
    """Dwass, Steel, Critchlow and Fligner all-pairs comparison test for a
    one-factorial layout with non-normally distributed residuals. As opposed to
    the all-pairs comparison procedures that depend on Kruskal ranks, the DSCF
    test is basically an extension of the U-test as re-ranking is conducted for
    each pairwise test [1]_, [2]_, [3]_.

    Parameters
    ----------
    a : Union[list, np.ndarray, DataFrame]
        An array, any object exposing the array interface or a pandas
        DataFrame.

    val_col : str, optional
        Name of a DataFrame column that contains dependent variable values (test
        or response variable). Values should have a non-nominal scale. Must be
        specified if `a` is a pandas DataFrame object.

    group_col : str, optional
        Name of a DataFrame column that contains independent variable values
        (grouping or predictor variable). Values should have a nominal scale
        (categorical). Must be specified if `a` is a pandas DataFrame object.

    sort : bool, optional
        If True, sort data by block and group columns.

    Returns
    -------
    result : pandas.DataFrame
        P values.

    Notes
    -----
    The p values are computed from the Tukey-distribution.

    References
    ----------
    .. [1] Douglas, C. E., Fligner, A. M. (1991) On distribution-free multiple
        comparisons in the one-way analysis of variance, Communications in
        Statistics - Theory and Methods, 20, 127-139.

    .. [2] Dwass, M. (1960) Some k-sample rank-order tests. In Contributions to
        Probability and Statistics, Edited by: I. Olkin, Stanford: Stanford
        University Press.

    .. [3] Steel, R. G. D. (1960) A rank sum test for comparing all pairs of
        treatments, Technometrics, 2, 197-207.

    Examples
    --------
    >>> import scikit_posthocs as sp
    >>> import pandas as pd
    >>> x = pd.DataFrame({"a": [1,2,3,5,1], "b": [12,31,54,62,12], "c": [10,12,6,74,11]})
    >>> x = x.melt(var_name='groups', value_name='values')
    >>> sp.posthoc_dscf(x, val_col='values', group_col='groups')
    """
    x, _val_col, _group_col = __convert_to_df(a, val_col, group_col)
    x = x.sort_values(by=[_group_col], ascending=True) if sort else x
    groups = x[_group_col].unique()
    x_grouped = x.groupby(_group_col, observed=True)[_val_col]
    n = x_grouped.count()
    k = groups.size

    def get_ties(x):
        t = x.value_counts().values
        c = np.sum((t**3 - t) / 12.0)
        return c

    def compare(i, j):
        ni = n.loc[i]
        nj = n.loc[j]
        x_raw = x.loc[(x[_group_col] == i) | (x[_group_col] == j)].copy()
        x_raw["ranks"] = x_raw.loc[:, _val_col].rank()
        r = x_raw.groupby(_group_col, observed=True)["ranks"].sum().loc[[j, i]]
        u = np.array([nj * ni + (nj * (nj + 1) / 2), nj * ni + (ni * (ni + 1) / 2)]) - r
        u_min = np.min(u)
        s = ni + nj
        var = (nj * ni / (s * (s - 1.0))) * ((s**3 - s) / 12.0 - get_ties(x_raw["ranks"]))
        p = np.sqrt(2.0) * (u_min - nj * ni / 2.0) / np.sqrt(var)
        return p

    vs = np.zeros((k, k))
    tri_upper = np.triu_indices(vs.shape[0], 1)
    tri_lower = np.tril_indices(vs.shape[0], -1)
    vs[:, :] = 0

    combs = it.combinations(range(k), 2)

    for i, j in combs:
        vs[i, j] = compare(groups[i], groups[j])

    vs[tri_upper] = ss.studentized_range.sf(np.abs(vs[tri_upper]), k, np.inf)
    vs[tri_lower] = np.transpose(vs)[tri_lower]

    np.fill_diagonal(vs, 1)
    return DataFrame(vs, index=groups, columns=groups)



from copy import deepcopy
from typing import Dict, List, Optional, Set, Tuple, Union

import numpy as np
from matplotlib import colors, pyplot
from matplotlib.axes import Axes
from matplotlib.colorbar import Colorbar, ColorbarBase
from matplotlib.colors import ListedColormap
from pandas import DataFrame, Index, Series
from seaborn import heatmap


def sign_array(p_values: Union[List, np.ndarray, DataFrame], alpha: float = 0.05) -> np.ndarray:
    """Significance array.

    Converts an array with p values to a significance array where
    0 is False (not significant), 1 is True (significant),
    and -1 is for diagonal elements.

    Parameters
    ----------
    p_values : Union[List, np.ndarray, DataFrame]
        Any object exposing the array interface and containing
        p values.

    alpha : float = 0.05
        Significance level. Default is 0.05.

    Returns
    -------
    result : numpy.ndarray
        Array where 0 is False (not significant), 1 is True (significant),
        and -1 is for diagonal elements.

    Examples
    --------
    >>> p_values = np.array([[ 1.        ,  0.00119517,  0.00278329],
                             [ 0.00119517,  1.        ,  0.18672227],
                             [ 0.00278329,  0.18672227,  1.        ]])
    >>> ph.sign_array(p_values)
    array([[-1,  1,  1],
           [ 1, -1,  0],
           [ 1,  0, -1]])
    """
    sig_array = deepcopy(np.array(p_values))
    sig_array[sig_array == 0] = 1e-10
    sig_array[sig_array > alpha] = 0
    sig_array[(sig_array < alpha) & (sig_array > 0)] = 1
    np.fill_diagonal(sig_array, -1)

    return sig_array


def sign_table(
    p_values: Union[List, np.ndarray, DataFrame], lower: bool = True, upper: bool = True
) -> Union[DataFrame, np.ndarray]:
    """Significance table.

    Returns table that can be used in a publication. P values are replaced
    with asterisks: \\* - p < 0.05, \\*\\* - p < 0.01, \\*\\*\\* - p < 0.001.

    Parameters
    ----------
    p_values : Union[List, np.ndarray, DataFrame]
        Any object exposing the array interface and containing
        p values.

    lower : bool
        Defines whether to return the lower triangle.

    upper : bool
        Defines whether to return the upper triangle.

    Returns
    -------
    result : Union[DataFrame, np.ndarray]
        P values masked with asterisks.

    Examples
    --------
    >>> p_values = np.array([[-1.        ,  0.00119517,  0.00278329],
                      [ 0.00119517, -1.        ,  0.18672227],
                      [ 0.00278329,  0.18672227, -1.        ]])
    >>> ph.sign_table(p_values)
    array([['-', '**', '**'],
           ['**', '-', 'NS'],
           ['**', 'NS', '-']], dtype=object)
    """
    if not any([lower, upper]):
        raise ValueError("Either lower or upper triangle must be returned")

    pv = DataFrame(p_values, copy=True) if not isinstance(p_values, DataFrame) else p_values.copy()

    ns = pv > 0.05
    three = (pv < 0.001) & (pv >= 0)
    two = (pv < 0.01) & (pv >= 0.001)
    one = (pv < 0.05) & (pv >= 0.01)

    pv = pv.astype(str)
    pv[ns] = "NS"
    pv[three] = "***"
    pv[two] = "**"
    pv[one] = "*"

    np.fill_diagonal(pv.values, "-")
    if not lower:
        pv.values[np.tril_indices(pv.shape[0], -1)] = ""
    elif not upper:
        pv.values[np.triu_indices(pv.shape[0], 1)] = ""

    return pv


def sign_plot(
    x: Union[List, np.ndarray, DataFrame],
    g: Union[List, np.ndarray, None] = None,
    flat: bool = False,
    labels: bool = True,
    cmap: Optional[List] = None,
    cbar_ax_bbox: Optional[Tuple[float, float, float, float]] = None,
    ax: Optional[Axes] = None,
    **kwargs,
) -> Union[Axes, Tuple[Axes, Colorbar]]:
    """Significance plot, a heatmap of p values (based on Seaborn).

    Parameters
    ----------
    x : Union[List, np.ndarray, DataFrame]
        If `flat` is False (default), `x` must be a square array, any object
        exposing the array interface, containing p values. If `flat` is True,
        `x` must be a sign_array
        (returned by :py:meth:`scikit_posthocs.sign_array` function).

    g : Union[List, np.ndarray]
        An array, any object exposing the array interface, containing
        group names.

    flat : bool
        If `flat` is True, plots a significance array as a heatmap using
        seaborn. If `flat` is False (default), plots an array of p values.
        Non-flat mode is useful if you need to  differentiate significance
        levels visually. It is the preferred mode.

    labels : bool
        Plot axes labels (default) or not.

    cmap : list
        1) If flat is False (default):
        List consisting of five elements, that will be exported to
        ListedColormap method of matplotlib. First is for diagonal
        elements, second is for non-significant elements, third is for
        p < 0.001, fourth is for p < 0.01, fifth is for p < 0.05.

        2) If flat is True:
        List consisting of three elements, that will be exported to
        ListedColormap method of matplotlib. First is for diagonal
        elements, second is for non-significant elements, third is for
        significant ones.
        3) If not defined, default colormaps will be used.

    cbar_ax_bbox : list
        Colorbar axes position rect [left, bottom, width, height] where
        all quantities are in fractions of figure width and height.
        Refer to `matplotlib.figure.Figure.add_axes` for more information.
        Default is [0.95, 0.35, 0.04, 0.3].

    ax : SubplotBase
        Axes in which to draw the plot, otherwise use the currently-active
        Axes.

    kwargs
        Keyword arguments to be passed to seaborn heatmap method. These
        keyword args cannot be used: cbar, vmin, vmax, center.

    Returns
    -------
    ax : matplotlib.axes._subplots.AxesSubplot
        Axes object with the heatmap.

    cbar : matplotlib.colorbar.Colorbar
        ColorBar object if `flat` is set to False.

    Examples
    --------
    >>> x = np.array([[ 1, 1, 1],
                      [ 1, 1, 0],
                      [ 1, 0, 1]])
    >>> ph.sign_plot(x, flat = True)
    """
    for key in ["cbar", "vmin", "vmax", "center"]:
        if key in kwargs:
            del kwargs[key]

    if isinstance(x, DataFrame):
        df = x.copy()
    else:
        g = g or np.arange(len(x))
        df = DataFrame(x, index=Index(g), columns=Index(g), copy=True)

    dtype = df.values.dtype

    if not np.issubdtype(dtype, np.integer) and flat:
        raise ValueError("X should be a sign_array or DataFrame of integers")
    elif not np.issubdtype(dtype, np.floating) and not flat:
        raise ValueError("X should be an array or DataFrame of float p values")

    if not cmap and flat:
        # format: diagonal, non-significant, significant
        cmap = ["1", "#fbd7d4", "#1a9641"]
    elif not cmap:
        # format: diagonal, non-significant, p<0.001, p<0.01, p<0.05
        cmap = ["1", "#fbd7d4", "#005a32", "#238b45", "#a1d99b"]

    if flat:
        np.fill_diagonal(df.values, -1)
        hax = heatmap(df, vmin=-1, vmax=1, cmap=ListedColormap(cmap), cbar=False, ax=ax, **kwargs)
        if not labels:
            hax.set_xlabel("")
            hax.set_ylabel("")
        return hax

    else:
        xc = df.values.copy()
        df[(xc < 0.001) & (xc >= 0)] = 1
        df[(xc < 0.01) & (xc >= 0.001)] = 2
        df[(xc < 0.05) & (xc >= 0.01)] = 3
        df[(xc >= 0.05)] = 0

        np.fill_diagonal(df.values, -1)

        if len(cmap) != 5:
            raise ValueError("Cmap list must contain 5 items")

        hax = heatmap(
            df,
            vmin=-1,
            vmax=3,
            cmap=ListedColormap(cmap),
            center=1,
            cbar=False,
            ax=ax,
            **kwargs,
        )
        if not labels:
            hax.set_xlabel("")
            hax.set_ylabel("")

        cbar_ax = hax.figure.add_axes(cbar_ax_bbox or (0.95, 0.35, 0.04, 0.3))
        cbar = ColorbarBase(
            cbar_ax,
            cmap=(ListedColormap(cmap[2:] + [cmap[1]])),
            norm=colors.NoNorm(),
            boundaries=[0, 1, 2, 3, 4],
        )
        cbar.set_ticks(
            list(np.linspace(0, 3, 4)),
            labels=["p < 0.001", "p < 0.01", "p < 0.05", "NS"],
        )

        cbar.outline.set_linewidth(1)
        cbar.outline.set_edgecolor("0.5")
        cbar.ax.tick_params(size=0)

        return hax, cbar


def _find_maximal_cliques(adj_matrix: DataFrame) -> List[Set]:
    """Wrapper function over the recursive Bron-Kerbosch algorithm.

    Will be used to find points that are under the same crossbar in critical
    difference diagrams.

    Parameters
    ----------
    adj_matrix : pandas.DataFrame
        Binary matrix with 1 if row item and column item do NOT significantly
        differ. Values in the main diagonal are not considered.

    Returns
    -------
    list[set]
        Largest fully connected subgraphs, represented as sets of indices of
        adj_matrix.

    Raises
    ------
    ValueError
        If the input matrix is empty or not symmetric.
        If the input matrix is not binary.

    """
    if (adj_matrix.index != adj_matrix.columns).any():
        raise ValueError("adj_matrix must be symmetric, indices do not match")
    if not adj_matrix.isin((0, 1)).values.all():
        raise ValueError("Input matrix must be binary")
    if adj_matrix.empty or not (adj_matrix.T == adj_matrix).values.all():
        raise ValueError("Input matrix must be non-empty and symmetric")

    result = []
    _bron_kerbosch(
        current_clique=set(),
        candidates=set(adj_matrix.index),
        visited=set(),
        adj_matrix=adj_matrix,
        result=result,
    )
    return result


def _bron_kerbosch(
    current_clique: Set,
    candidates: Set,
    visited: Set,
    adj_matrix: DataFrame,
    result: List[Set],
) -> None:
    """Recursive algorithm to find the maximal fully connected subgraphs.

    See [1]_ for more information.

    Parameters
    ----------
    current_clique : set
        A set of vertices known to be fully connected.
    candidates : set
        Set of vertices that could potentially be added to the clique.
    visited : set
        Set of vertices already known to be part of another previously explored
        clique, that is not current_clique.
    adj_matrix : pandas.DataFrame
        Binary matrix with 1 if row item and column item do NOT significantly
        differ. Diagonal must be zeroed.
    result : list[set]
        List where to append the maximal cliques.

    Returns
    -------
    None

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Bron%E2%80%93Kerbosch_algorithm
    """
    while candidates:
        v = candidates.pop()
        _bron_kerbosch(
            current_clique | {v},
            # Restrict candidate vertices to the neighbors of v
            {n for n in candidates if adj_matrix.loc[v, n]},
            # Restrict visited vertices to the neighbors of v
            {n for n in visited if adj_matrix.loc[v, n]},
            adj_matrix,
            result,
        )
        visited.add(v)

    # We do not need to report a clique if a children call aready did it.
    if not visited:
        # If this is not a terminal call, i.e. if any clique was reported.
        result.append(current_clique)


def critical_difference_diagram(
    ranks: Union[dict, Series],
    sig_matrix: DataFrame,
    *,
    ax: Optional[Axes] = None,
    label_fmt_left: str = "{label} ({rank:.2g})",
    label_fmt_right: str = "({rank:.2g}) {label}",
    label_props: Optional[dict] = None,
    marker_props: Optional[dict] = None,
    elbow_props: Optional[dict] = None,
    crossbar_props: Optional[dict] = None,
    color_palette: Union[Dict[str, str], List, None] = None,
    text_h_margin: float = 0.01,
    left_only: bool = False,
) -> Dict[str, list]:
    """Plot a Critical Difference diagram from ranks and post-hoc results.

    The diagram arranges the average ranks of multiple groups on the x axis
    in order to facilitate performance comparisons between them. The groups
    that could not be statistically deemed as different are linked by a
    horizontal crossbar [1]_, [2]_.

    ::

                      rank markers
         X axis ---------O----O-------------------O-O------------O---------
                         |----|                   | |            |
                         |    |                   |---crossbar---|
                clf1 ----|    |                   | |            |---- clf3
                clf2 ---------|                   | |----------------- clf4
                                                  |------------------- clf5
                    |____|
                text_h_margin

    In the drawing above, the two crossbars indicate that clf1 and clf2 cannot
    be statistically differentiated, the same occurring between clf3, clf4 and
    clf5. However, clf1 and clf2 are each significantly lower ranked than clf3,
    clf4 and clf5.

    Parameters
    ----------
    ranks : dict or Series
        Indicates the rank value for each sample or estimator (as keys or index).

    sig_matrix : DataFrame
        The corresponding p-value matrix outputted by post-hoc tests, with
        indices matching the labels in the ranks argument.

    ax : matplotlib.SubplotBase, optional
        The object in which the plot will be built. Gets the current Axes
        by default (if None is passed).

    label_fmt_left : str, optional
        The format string to apply to the labels on the left side. The keywords
        label and rank can be used to specify the sample/estimator name and
        rank value, respectively, by default '{label} ({rank:.2g})'.

    label_fmt_right : str, optional
        The same, but for the labels on the right side of the plot.
        By default '({rank:.2g}) {label}'.

    label_props : dict, optional
        Parameters to be passed to pyplot.text() when creating the labels,
        by default None.

    marker_props : dict, optional
        Parameters to be passed to pyplot.scatter() when plotting the rank
        markers on the axis, by default None.

    elbow_props : dict, optional
        Parameters to be passed to pyplot.plot() when creating the elbow lines,
        by default None.

    crossbar_props : dict, optional
        Parameters to be passed to pyplot.plot() when creating the crossbars
        that indicate lack of statistically significant difference. By default
        None.

    color_palette: dict or list, optional
        Parameters to be passed when you need specific colors for each category

    text_h_margin : float, optional
        Space between the text labels and the nearest vertical line of an
        elbow, by default 0.01.

    left_only: boolean, optional
        Set all labels in a single left-sided block instead of splitting them
        into two block, one for the left and one for the right.


    Returns
    -------
    dict[str, list[matplotlib.Artist]]
        Lists of Artists created.

    Examples
    --------
    See the :doc:`/tutorial`.

    References
    ----------
    .. [1] Demšar, J. (2006). Statistical comparisons of classifiers over multiple
            data sets. The Journal of Machine learning research, 7, 1-30.

    .. [2] https://mirkobunse.github.io/CriticalDifferenceDiagrams.jl/stable/
    """
    ## check color_palette consistency
    if not color_palette or len(color_palette) == 0:
        pass
    elif isinstance(color_palette, Dict) and (
        (len(set(ranks.keys()) & set(color_palette.keys()))) == len(ranks)
    ):
        pass
    elif isinstance(color_palette, List) and (len(ranks) <= len(color_palette)):
        pass
    else:
        raise ValueError("color_palette keys are not consistent, or list size too small")

    elbow_props = elbow_props or {}
    marker_props = {"zorder": 3, **(marker_props or {})}
    label_props = {"va": "center", **(label_props or {})}
    crossbar_props = {
        "color": "k",
        "zorder": 3,
        "linewidth": 2,
        **(crossbar_props or {}),
    }

    ax = ax or pyplot.gca()
    ax.yaxis.set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["left"].set_visible(False)
    ax.spines["bottom"].set_visible(False)
    ax.xaxis.set_ticks_position("top")
    ax.spines["top"].set_position("zero")

    # lists of artists to be returned
    markers = []
    elbows = []
    labels = []
    crossbars = []

    # True if pairwise comparison is NOT significant
    adj_matrix = DataFrame(
        1 - sign_array(sig_matrix),
        index=sig_matrix.index,
        columns=sig_matrix.columns,
        dtype=bool,
    )

    ranks = Series(ranks).sort_values()  # Standardize if ranks is dict
    if left_only:
        points_left = ranks
    else:
        points_left, points_right = (
            ranks.iloc[: len(ranks) // 2],
            ranks.iloc[len(ranks) // 2 :],
        )
    # points_left, points_right = np.array_split(ranks.sort_values(), 2)

    # Sets of points under the same crossbar
    crossbar_sets = _find_maximal_cliques(adj_matrix)

    # Sort by lowest rank and filter single-valued sets
    crossbar_sets = sorted(
        (x for x in crossbar_sets if len(x) > 1), key=lambda x: ranks[list(x)].min()
    )

    # Create stacking of crossbars: for each level, try to fit the crossbar,
    # so that it does not intersect with any other in the level. If it does not
    # fit in any level, create a new level for it.
    crossbar_levels: list[list[set]] = []
    for bar in crossbar_sets:
        for level, bars_in_level in enumerate(crossbar_levels):
            if not any(bool(bar & bar_in_lvl) for bar_in_lvl in bars_in_level):
                ypos = -level - 1
                bars_in_level.append(bar)
                break
        else:
            ypos = -len(crossbar_levels) - 1
            crossbar_levels.append([bar])

        crossbars.append(
            ax.plot(
                # Adding a separate line between each pair enables showing a
                # marker over each elbow with crossbar_props={'marker': 'o'}.
                [ranks[i] for i in bar],
                [ypos] * len(bar),
                **crossbar_props,
            )
        )

    lowest_crossbar_ypos = -len(crossbar_levels)

    def plot_items(points, xpos, label_fmt, color_palette, label_props):
        """Plot each marker + elbow + label."""
        ypos = lowest_crossbar_ypos - 1
        for idx, (label, rank) in enumerate(points.items()):
            if not color_palette or len(color_palette) == 0:
                elbow, *_ = ax.plot(
                    [xpos, rank, rank],
                    [ypos, ypos, 0],
                    **elbow_props,
                )
            else:
                elbow, *_ = ax.plot(
                    [xpos, rank, rank],
                    [ypos, ypos, 0],
                    c=color_palette[label]
                    if isinstance(color_palette, Dict)
                    else color_palette[idx],
                    **elbow_props,
                )

            elbows.append(elbow)
            curr_color = elbow.get_color()
            markers.append(ax.scatter(rank, 0, **{"color": curr_color, **marker_props}))
            labels.append(
                ax.text(
                    xpos,
                    ypos,
                    label_fmt.format(label=label, rank=rank),
                    color=curr_color,
                    **label_props,
                )
            )
            ypos -= 1

    plot_items(
        points_left,
        xpos=points_left.iloc[0] - text_h_margin,
        label_fmt=label_fmt_left,
        color_palette=color_palette,
        label_props={
            "ha": "right",
            **label_props,
        },
    )

    if not left_only:
        plot_items(
            points_right[::-1],
            xpos=points_right.iloc[-1] + text_h_margin,
            label_fmt=label_fmt_right,
            color_palette=list(reversed(color_palette))
            if isinstance(color_palette, list)
            else color_palette,
            label_props={"ha": "left", **label_props},
        )

    return {
        "markers": markers,
        "elbows": elbows,
        "labels": labels,
        "crossbars": crossbars,
    }



# -*- coding: utf-8 -*-

from typing import Optional, Union, List, cast
import itertools as it
import numpy as np
from numpy.typing import ArrayLike
import scipy.stats as ss
from pandas import DataFrame, Categorical, Series
from scikit_posthocs._posthocs import __convert_to_df, __convert_to_block_df


def test_mackwolfe(
    data: Union[ArrayLike, DataFrame],
    val_col: Optional[str] = None,
    group_col: Optional[str] = None,
    p: Optional[int] = None,
    n_perm: int = 100,
    sort: bool = False,
) -> tuple[float, float]:
    """Mack-Wolfe Test for Umbrella Alternatives.

    In dose-finding studies one may assume an increasing treatment effect with
    increasing dose level. However, the test subject may actually succumb to
    toxic effects at high doses, which leads to decresing treatment
    effects [1]_, [2]_.

    The scope of the Mack-Wolfe Test is to test for umbrella alternatives for
    either a known or unknown point P (i.e. dose-level), where the peak
    (umbrella point) is present.

    Parameters
    ----------
    data : Union[List, numpy.ndarray, DataFrame]
        An array, any object exposing the array interface or a pandas
        DataFrame with data values.

    val_col : str = None
        Name of a DataFrame column that contains dependent variable values
        (test or response variable). Values should have a non-nominal scale.
        Must be specified if ``a`` is a pandas DataFrame object.

    group_col : str = None
        Name of a DataFrame column that contains independent variable values
        (grouping or predictor variable). Values should have a nominal scale
        (categorical). Must be specified if ``a`` is a pandas DataFrame object.

    p : int = None
        The a priori known peak as an ordinal number of the treatment group
        including the zero dose level, i.e. p = {0, ..., k-1}.
        Defaults to None.

    n_perm: int = 100
        Permutations number.

    sort : bool = False
        If ``True``, sort data by block and group columns.

    Returns
    -------
    tuple[float, float]
        P value and statistic.

    References
    ----------
    .. [1] Chen, I.Y. (1991) Notes on the Mack-Wolfe and Chen-Wolfe Tests for
        Umbrella Alternatives. Biom. J., 33, 281-290.
    .. [2] Mack, G.A., Wolfe, D. A. (1981) K-sample rank tests for umbrella
        alternatives. J. Amer. Statist. Assoc., 76, 175-181.

    Examples
    --------
    >>> x = [[22, 23, 35], [60, 59, 54], [98, 78, 50], [60, 82, 59], [22, 44, 33], [23, 21, 25]]
    >>> sp.posthoc_mackwolfe(x)
    """
    x, _val_col, _group_col = __convert_to_df(data, val_col, group_col)

    if not sort:
        x[_group_col] = Categorical(x[_group_col], categories=x[_group_col].unique(), ordered=True)
    x.sort_values(by=[_group_col], ascending=True, inplace=True)

    k = x[_group_col].unique().size

    if p and p > k:
        print("Selected 'p' > number of groups:", str(p), " > ", str(k))
        return (np.nan, np.nan)
    elif p is not None and p < 1:
        print("Selected 'p' < 1: ", str(p))
        return (np.nan, np.nan)

    Rij = x[_val_col].rank()
    n = cast(Series, x.groupby(_group_col, observed=True)[_val_col].count())

    def _fn(Ri, Rj):
        return np.sum(Ri.apply(lambda x: Rj[Rj > x].size))

    def _ustat(Rij, g, k):
        levels = np.unique(g)
        U = np.identity(k)

        for i in range(k):
            for j in range(i):
                U[i, j] = _fn(Rij[x[_group_col] == levels[i]], Rij[x[_group_col] == levels[j]])
                U[j, i] = _fn(Rij[x[_group_col] == levels[j]], Rij[x[_group_col] == levels[i]])

        return U

    def _ap(p, U) -> float:
        tmp1 = 0.0
        if p > 0:
            for i in range(p):
                for j in range(i + 1, p + 1):
                    tmp1 += U[i, j]
        tmp2 = 0.0
        if p < k:
            for i in range(p, k):
                for j in range(i + 1, k):
                    tmp2 += U[j, i]

        return tmp1 + tmp2

    def _n1(p: int, n: Series) -> float:
        return np.sum(n[: p + 1])

    def _n2(p: int, n: Series) -> float:
        return np.sum(n[p:k])

    def _mean_at(p, n) -> float:
        N1 = _n1(p, n)
        N2 = _n2(p, n)
        return (N1**2.0 + N2**2.0 - np.sum(n**2.0) - n.iloc[p] ** 2.0) / 4.0

    def _var_at(p: int, n: Series) -> float:
        N1 = _n1(p, n)
        N2 = _n2(p, n)
        N = np.sum(n)

        var = (
            2.0 * (N1**3 + N2**3)
            + 3.0 * (N1**2 + N2**2)
            - np.sum(n**2 * (2 * n + 3.0))
            - n.iloc[p] ** 2.0 * (2.0 * n.iloc[p] + 3.0)
            + 12.0 * n.iloc[p] * N1 * N2
            - 12.0 * n.iloc[p] ** 2.0 * N
        ) / 72.0
        return var

    if p:
        # if (x.groupby(_val_col).count() > 1).any().any():
        #    print("Ties are present")
        U = _ustat(Rij, x[_group_col], k)
        est = _ap(p, U)
        mean = _mean_at(p, n)
        sd = np.sqrt(_var_at(p, n))
        stat = (est - mean) / sd
        p_value = ss.norm.sf(stat).item()
    else:
        U = _ustat(Rij, x[_group_col], k)
        Ap = np.array([_ap(i, U) for i in range(k)]).ravel()
        mean = np.array([_mean_at(i, n) for i in range(k)]).ravel()
        var = np.array([_var_at(i, n) for i in range(k)]).ravel()
        A = (Ap - mean) / np.sqrt(var)
        stat = float(np.max(A))

        mt = []
        for _ in range(n_perm):
            ix = Series(np.random.permutation(Rij))
            uix = _ustat(ix, x[_group_col], k)
            apix = np.array([_ap(i, uix) for i in range(k)])
            astarix = (apix - mean) / np.sqrt(var)
            mt.append(np.max(astarix))

        mt = np.array(mt)
        p_value = mt[mt > stat].size / n_perm

    return p_value, stat


def test_osrt(
    data: Union[List, np.ndarray, DataFrame],
    val_col: Optional[str] = None,
    group_col: Optional[str] = None,
    sort: bool = False,
) -> tuple[float, float, int]:
    """Hayter's one-sided studentised range test (OSRT)

    Tests a hypothesis against an ordered alternative for normal data with
    equal variances [1]_.

    Parameters
    ----------
    data : Union[List, numpy.ndarray, DataFrame]
        An array, any object exposing the array interface or a pandas
        DataFrame with data values.

    val_col : str = None
        Name of a DataFrame column that contains dependent variable values
        (test or response variable). Values should have a non-nominal scale.
        Must be specified if ``a`` is a pandas DataFrame object.

    group_col : str = None
        Name of a DataFrame column that contains independent variable values
        (grouping or predictor variable). Values should have a nominal scale
        (categorical). Must be specified if `a` is a pandas DataFrame object.

    sort : bool = False
        If True, sort data by block and group columns.

    Returns
    -------
    tuple[float, float, int]
        P value, statistic, and number of degrees of freedom.

    Notes
    -----
    P values are computed from the Tukey distribution.

    References
    ----------
    .. [1] Hayter, A.J.(1990) A One-Sided Studentised Range Test for Testing
        Against a Simple Ordered Alternative, Journal of the American
        Statistical Association, 85, 778-785.

    Examples
    --------
    >>> import scikit_posthocs as sp
    >>> import pandas as pd
    >>> x = pd.DataFrame({"a": [1,2,3,5,1], "b": [12,31,54,62,12], "c": [10,12,6,74,11]})
    >>> x = x.melt(var_name='groups', value_name='values')
    >>> sp.test_osrt(x, val_col='values', group_col='groups')
    """
    x, _val_col, _group_col = __convert_to_df(data, val_col, group_col)

    if not sort:
        x[_group_col] = Categorical(x[_group_col], categories=x[_group_col].unique(), ordered=True)

    x.sort_values(by=[_group_col], ascending=True, inplace=True)
    groups = np.unique(x[_group_col])
    x_grouped = x.groupby(_group_col, observed=True)[_val_col]

    xi = x_grouped.mean()
    ni = x_grouped.count()
    k = groups.size
    n = len(x.index)
    df = n - k

    sigma2 = 0
    c = -1

    for i in range(k):
        for j in range(ni.iloc[i]):
            c += 1
            sigma2 += (x[_val_col].iloc[c] - xi.iloc[i]) ** 2.0 / df

    sigma = np.sqrt(sigma2)

    def compare(i, j):
        dif = xi.loc[groups[j]] - xi.loc[groups[i]]
        A = sigma / np.sqrt(2.0) * np.sqrt(1.0 / ni[groups[j]] + 1.0 / ni[groups[i]])
        qval = np.abs(dif) / A
        return qval

    vs = np.zeros((k, k), dtype=float)
    combs = it.combinations(range(k), 2)

    for i, j in combs:
        vs[i, j] = compare(i, j)

    stat = np.max(vs)
    pval = ss.studentized_range.sf(stat, k, df)
    return pval, stat, df


def test_durbin(
    data: Union[List, np.ndarray, DataFrame],
    y_col: Optional[Union[str, int]] = None,
    group_col: Optional[Union[str, int]] = None,
    block_col: Optional[Union[str, int]] = None,
    block_id_col: Optional[Union[str, int]] = None,
    melted: bool = False,
    sort: bool = True,
) -> tuple[float, float, int]:
    """Durbin's test whether k groups (or treatments) in a two-way
    balanced incomplete block design (BIBD) have identical effects. See
    references for additional information [1]_, [2]_.

    Parameters
    ----------
    data : Union[List, np.ndarray, DataFrame]
        An array, any object exposing the array interface or a pandas
        DataFrame with data values.

        If ``melted`` argument is set to False (default), ``a`` is a typical
        matrix of block design, i.e. rows are blocks, and columns are groups.
        In this case, you do not need to specify col arguments.

        If ``a`` is an array and ``melted`` is set to True,
        y_col, block_col and group_col must specify the indices of columns
        containing elements of correspondary type.

        If ``a`` is a Pandas DataFrame and ``melted`` is set to True,
        y_col, block_col and group_col must specify columns names (string).

    y_col : Union[str, int] = None
        Must be specified if ``a`` is a melted pandas DataFrame object.
        Name of the column that contains y data.

    group_col : Union[str, int] = None
        Must be specified if ``a`` is a melted pandas DataFrame object.
        Name of the column that contains group names.

    block_col : Union[str, int] = None
        Must be specified if ``a`` is a melted pandas DataFrame object.
        Name of the column that contains block names.

    block_id_col : Union[str, int] = None
        Must be specified if ``a`` is a melted pandas DataFrame object.
        Name of the column that contains identifiers of block names.
        In most cases, this is the same as `block_col` except for those
        cases when you have multiple instances of the same blocks.

    melted : bool = False
        Specifies if data are given as melted columns "y", "blocks", and
        "groups".

    sort : bool = False
        If True, sort data by block and group columns.

    Returns
    -------
    tuple[float, float, int]
        P value, statistic, and number of degrees of freedom.

    References
    ----------
    .. [1] N. A. Heckert, J. J. Filliben. (2003) NIST Handbook 148: Dataplot Reference
        Manual, Volume 2: Let Subcommands and Library Functions. National Institute of
        Standards and Technology Handbook Series, June 2003.
    .. [2] W. J. Conover (1999), Practical nonparametric Statistics,
        3rd. edition, Wiley.

    Examples
    --------
    >>> x = np.array([[31,27,24],[31,28,31],[45,29,46],[21,18,48],[42,36,46],[32,17,40]])
    >>> sp.test_durbin(x)
    """
    x, _y_col, _group_col, _block_col, _block_id_col = __convert_to_block_df(
        data, y_col, group_col, block_col, block_id_col, melted
    )

    groups = x[_group_col].unique()
    blocks = x[_block_id_col].unique()
    if not sort:
        x[_group_col] = Categorical(x[_group_col], categories=groups, ordered=True)
        x[_block_col] = Categorical(x[_block_col], categories=blocks, ordered=True)
    x.sort_values(by=[_block_col, _group_col], ascending=True, inplace=True)
    x.dropna(inplace=True)

    t = len(groups)
    b = len(blocks)
    r = float(b)
    k = float(t)

    x["y_ranks"] = x.groupby(_block_id_col, observed=True)[_y_col].rank()
    rs = x.groupby(_group_col, observed=True)["y_ranks"].sum().to_numpy()

    A = float(np.sum(x["y_ranks"] ** 2.0))
    C = float(b * k * (k + 1) ** 2.0) / 4.0
    D = float(np.sum(rs**2.0)) - r * C
    T1 = (t - 1.0) / (A - C) * D
    stat = T1
    df = t - 1
    pval = ss.chi2.sf(stat, df).item()

    return pval, stat, df
